#include "CoordinateSystem3D.h"

CoordinateSystem3D::CoordinateSystem3D(const String &name) 
 :	Component(name)
{
 scaleX                        =  1.0;
 scaleY                        =  1.0;
 scaleZ                        =  1.0;

 pixelsPerIntervalX            =  100.0;
 pixelsPerIntervalY            =  100.0;
 pixelsPerIntervalZ            =  100.0;

 axisPositionX                 =  ZERO;
 axisPositionY                 =  ZERO;
 axisPositionZ                 =  ZERO;

 axisLabelX                    =  String(T("x"));
 axisLabelY                    =  String(T("y"));
 axisLabelZ                    =  String(T("z"));

 axisLabelPositionX            =  ABOVE_AXIS;
 axisLabelPositionY            =  RIGHT_TO_AXIS;
 axisLabelPositionZ            =  RIGHT_TO_AXIS;

 axisValuesPositionX           =  BELOW_AXIS;
 axisValuesPositionY           =  LEFT_TO_AXIS;
 axisValuesPositionZ           =  LEFT_TO_AXIS;

 coarseGridIsVisibleX          = true;
 fineGridIsVisibleX            = false;
 coarseGridIsVisibleY          = true;
 fineGridIsVisibleY            = false;
 coarseGridIsVisibleZ          = true;
 fineGridIsVisibleZ            = false;

 coarseGridIntervalX           =  1.0;
 fineGridIntervalX             =  0.1;
 coarseGridIntervalY           =  1.0;
 fineGridIntervalY             =  0.1;
 coarseGridIntervalZ           =  1.0;
 fineGridIntervalZ             =  0.1;

 isLogScaledX	                 =  false;
 isLogScaledY	                 =  false;
 isLogScaledZ	                 =  false;

 logBaseX	                     =  2.0;
 logBaseY	                     =  2.0;
 logBaseZ	                     =  2.0;


 projectionMethod              = AXONOMETRIC;

 rotationAboutY                = -45.0; 
 rotationAboutX                = asin(sqrt(1.0/3.0)) * 180.0/M_PI; 
 rotationAboutZ                = 0.0; 

 horizontalScale               = 0.57;
 verticalScale                 = 0.57;
 horizontalShift               = 0.0;
 verticalShift                 = 0.0;

 maxScale                      = 10.0;
 minScale                      =  0.1;
 maxShift                      =  1.0;
 minShift                      = -1.0;

 // initialize the transformation (matrix-) parameters:
 a = 1.0; b = 0.0; c = 0.0; p = 0.0;
 d = 0.0; e = 1.0; f = 0.0; q = 0.0;
 h = 0.0; i = 0.0; j = 1.0; r = 0.0;
 l = 0.0; m = 0.0; n = 0.0; s = 1.0;

 // initialize the function-pointers for value->string conversion
 stringConversionFunctionX = &valueToString0;
 stringConversionFunctionY = &valueToString0;
 stringConversionFunctionZ = &valueToString0;

 // set up the default colour-scheme:
 backgroundColour              = Colours::white;
 axesColour                    = Colours::black;
 coarseGridColour              = Colours::grey.withAlpha(0.5f);
 fineGridColour                = Colours::grey.withAlpha(0.125f);
 clippingPlaneColourXY         = Colours::red.withAlpha(0.125f);
 clippingPlaneColourXZ         = Colours::green.withAlpha(0.125f);
 clippingPlaneColourYZ         = Colours::blue.withAlpha(0.125f);
 clippingPlanesOutlineColour   = coarseGridColour;
 outlineColour                 = Colours::black;

 // initialize the component-size and the image-size to 1x1 pixels, without
 // such initializations, a JUCE-breakpoint will be triggered or other screws
 // happen:
 backgroundImage = NULL;
 backgroundImage = new Image(Image::RGB, 1, 1, true);
 setBounds(0, 0, 1, 1);
 updateBackgroundImage();
}

CoordinateSystem3D::~CoordinateSystem3D()
{
 if( backgroundImage != NULL)
  delete backgroundImage;
}

//----------------------------------------------------------------------------
void CoordinateSystem3D::setBounds(int x, int y, int width, int height)
{
 Component::setBounds(x, y, width, height);

 updateProjectionParameters();
 updateBackgroundImage();
}

//----------------------------------------------------------------------------

void CoordinateSystem3D::setAxisLabel(const String& newLabel, int whichAxis,
                                      int newLabelPosition)
{
 switch( whichAxis )
 {
 case X_AXIS:
  {
   axisLabelX         = newLabel;
   axisLabelPositionX = newLabelPosition;
  } 
  break;
 case Y_AXIS:
  {
   axisLabelY         = newLabel;
   axisLabelPositionY = newLabelPosition;
  } 
  break;
 case Z_AXIS:
  {
   axisLabelZ         = newLabel;
   axisLabelPositionZ = newLabelPosition;
  } 
  break;
 }
 updateBackgroundImage();
}

void CoordinateSystem3D::setValuesPosition(int newValuesPosition, 
                                           int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS:
  {
   if( newValuesPosition == NO_ANNOTATION ||
       newValuesPosition == BELOW_AXIS  ||
       newValuesPosition == ABOVE_AXIS   )
   {
    axisValuesPositionX = newValuesPosition;
   }
  } 
  break;
 case Y_AXIS:
  {
   if( newValuesPosition == NO_ANNOTATION ||
       newValuesPosition == LEFT_TO_AXIS  ||
       newValuesPosition == RIGHT_TO_AXIS   )
   {
    axisValuesPositionY = newValuesPosition;
   }
  } 
  break;
 case Z_AXIS:
  {
   if( newValuesPosition == NO_ANNOTATION ||
       newValuesPosition == LEFT_TO_AXIS  ||
       newValuesPosition == RIGHT_TO_AXIS   )
   {
    axisValuesPositionZ = newValuesPosition;
   }
  } 
  break;
 }

 updateBackgroundImage();
}

void CoordinateSystem3D::setStringConversionFunction(
 String (*newConversionFunction) (double valueToBeConverted), int whichAxis )
{
 switch( whichAxis )
 {
 case X_AXIS:
  stringConversionFunctionX = newConversionFunction;
  break;
 case Y_AXIS:
  stringConversionFunctionY = newConversionFunction;
  break;
 case Z_AXIS:
  stringConversionFunctionZ = newConversionFunction;
  break;
 }
}

void CoordinateSystem3D::setColourScheme(Colour newBackgroundColour, 
                                         Colour newAxesColour,
                                         Colour newCoarseGridColour,
                                         Colour newFineGridColour,
                                         Colour newOutlineColour)
{
 backgroundColour = newBackgroundColour;
 axesColour       = newAxesColour;
 coarseGridColour = newCoarseGridColour;
 fineGridColour   = newFineGridColour;
 outlineColour    = newOutlineColour;

 updateBackgroundImage();
}

void CoordinateSystem3D::setHorizontalScaleFactor(double newScaleFactor)
{
 horizontalScale = newScaleFactor;
 if( horizontalScale > maxScale )
  horizontalScale = maxScale;
 else if ( horizontalScale < minScale )
  horizontalScale = minScale;

 updateBackgroundImage();
}

double CoordinateSystem3D::getHorizontalScaleFactor()
{
 return horizontalScale;
}

void CoordinateSystem3D::setVerticalScaleFactor(double newScaleFactor)
{
 verticalScale = newScaleFactor;
 if( verticalScale > maxScale )
  verticalScale = maxScale;
 else if ( verticalScale < minScale )
  verticalScale = minScale;

 updateBackgroundImage();
}

double CoordinateSystem3D::getVerticalScaleFactor()
{
 return verticalScale;
}

void CoordinateSystem3D::setHorizontalShift(double newShift)
{
 horizontalShift = newShift;
 if( horizontalShift > maxShift )
  horizontalShift = maxShift;
 else if ( horizontalShift < minShift )
  horizontalShift = minShift;

 updateBackgroundImage();
}

double CoordinateSystem3D::getHorizontalShift()
{
 return horizontalShift;
}

void CoordinateSystem3D::setVerticalShift(double newShift)
{
 verticalShift = newShift;
 if( verticalShift > maxShift )
  verticalShift = maxShift;
 else if ( verticalShift < minShift )
  verticalShift = minShift;

 updateBackgroundImage();
}

double CoordinateSystem3D::getVerticalShift()
{
 return verticalShift;
}

void CoordinateSystem3D::scaleHorizontalScaleFactor(double scaler)
{
 horizontalScale *= scaler;
 if( horizontalScale > maxScale )
  horizontalScale = maxScale;
 else if ( horizontalScale < minScale )
  horizontalScale = minScale;

 updateBackgroundImage();
}

void CoordinateSystem3D::scaleVerticalScaleFactor(double scaler)
{
 verticalScale *= scaler;
 if( verticalScale > maxScale )
  verticalScale = maxScale;
 else if ( verticalScale < minScale )
  verticalScale = minScale;

 updateBackgroundImage();
}

void CoordinateSystem3D::incrementHorizontalShift(double increment)
{
 horizontalShift += increment;
 if( horizontalShift > maxShift )
  horizontalShift = maxShift;
 else if ( horizontalShift < minShift )
  horizontalShift = minShift;

 updateBackgroundImage();
}

void CoordinateSystem3D::incrementVerticalShift(double increment)
{
 verticalShift += increment;
 if( verticalShift > maxShift )
  verticalShift = maxShift;
 else if ( verticalShift < minShift )
  verticalShift = minShift;

 updateBackgroundImage();
}

void CoordinateSystem3D::incrementRotationAngle(double increment, 
                                                int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS: 
  rotationAboutX = fmod(rotationAboutX+increment, 360.0); 
  break;
 case Y_AXIS: 
  rotationAboutY = fmod(rotationAboutY+increment, 360.0); 
  break;
 case Z_AXIS: 
  rotationAboutZ = fmod(rotationAboutZ+increment, 360.0); 
  break;
 }

 updateProjectionParameters();
 updateBackgroundImage();
}


void CoordinateSystem3D::setRange(double newMinX, double newMaxX, 
			                            	  double newMinY, double newMaxY,
                                  double newMinZ, double newMaxZ)
{
 range3D.setMinX(newMinX);
 range3D.setMaxX(newMaxX);
 range3D.setMinY(newMinY);
 range3D.setMaxY(newMaxY);
 range3D.setMinZ(newMinZ);
 range3D.setMaxZ(newMaxZ);

 updateProjectionParameters();
 updateBackgroundImage();
}

void CoordinateSystem3D::setRange(CoordinateSystem3DRange newRange)
{
 range3D = newRange;
 updateProjectionParameters();
 updateBackgroundImage();
}

void CoordinateSystem3D::setMinimum(double newMinimum, int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS: range3D.setMinX(newMinimum); break;
 case Y_AXIS: range3D.setMinY(newMinimum); break; 
 case Z_AXIS: range3D.setMinZ(newMinimum); break;
 }

 updateProjectionParameters();
 updateBackgroundImage();
}

double CoordinateSystem3D::getMinimum(int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS: return range3D.getMinX(); 
 case Y_AXIS: return range3D.getMinY(); 
 case Z_AXIS: return range3D.getMinZ(); 
 default:     return 0.0;
 }
}

void CoordinateSystem3D::setMaximum(double newMaximum, int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS: range3D.setMaxX(newMaximum); break; 
 case Y_AXIS: range3D.setMaxY(newMaximum); break; 
 case Z_AXIS: range3D.setMaxZ(newMaximum); break; 
 }

 updateProjectionParameters();
 updateBackgroundImage();
}

double CoordinateSystem3D::getMaximum(int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS: return range3D.getMaxX(); 
 case Y_AXIS: return range3D.getMaxY(); 
 case Z_AXIS: return range3D.getMaxZ(); 
 default:     return 0.0;
 }
}

CoordinateSystem3DRange // return value
 CoordinateSystem3D::getMaximumMeaningfulRange3D(double relativeMargin)
{
 // require at least 1.0% margin:
 jassert( relativeMargin >= 1.0 );
 if( relativeMargin < 1.0 )
  relativeMargin = 1.0;

 CoordinateSystem3DRange r;

 /*  TODO: do something more here
 double minX   = range3D.getMinX();
 double maxX   = range3D.getMaxX();
 double width  = maxX-minX;
 double minY   = range3D.getMinY();
 double maxY   = range3D.getMaxY();
 double height = maxY-minY;

 r.setMinX(minX - 0.01*relativeMargin*width);
 r.setMaxX(maxX + 0.01*relativeMargin*width);
 r.setMinY(minY - 0.01*relativeMargin*height);
 r.setMaxY(maxY + 0.01*relativeMargin*height);
 */

 return r;
}

void CoordinateSystem3D::setCoarseGrid(double newGridInterval, 
                                       bool   shouldBeVisible, int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS:
  {
   if( isLogScaledX )
   {
    jassert(newGridInterval > 1.00001);
     // for logarithmic scaling, we need the grid-intervals to be > 1
    if( newGridInterval <= 1.00001 )
    {
     coarseGridIntervalX = 2.0;
     return;
    }
   }
   else
   {
    jassert(newGridInterval > 0.000001); 
     // grid-intervals must be > 0
    if( newGridInterval <= 0.000001 )
     return;
   }
   coarseGridIsVisibleX = shouldBeVisible;
   coarseGridIntervalX  = newGridInterval;
  } 
  break;
 case Y_AXIS:
  {
   if( isLogScaledY )
   {
    jassert(newGridInterval > 1.00001);
     // for logarithmic scaling, we need the grid-intervals to be > 1
    if( newGridInterval <= 1.00001 )
    {
     coarseGridIntervalY = 2.0;
     return;
    }
   }
   else
   {
    jassert(newGridInterval > 0.000001); 
     // grid-intervals must be > 0
    if( newGridInterval <= 0.000001 )
     return;
   }
   coarseGridIsVisibleY = shouldBeVisible;
   coarseGridIntervalY  = newGridInterval;
  } 
  break;
 case Z_AXIS:
  {
   if( isLogScaledZ )
   {
    jassert(newGridInterval > 1.00001);
     // for logarithmic scaling, we need the grid-intervals to be > 1
    if( newGridInterval <= 1.00001 )
    {
     coarseGridIntervalZ = 2.0;
     return;
    }
   }
   else
   {
    jassert(newGridInterval > 0.000001); 
     // grid-intervals must be > 0
    if( newGridInterval <= 0.000001 )
     return;
   }
   coarseGridIsVisibleZ = shouldBeVisible;
   coarseGridIntervalZ  = newGridInterval;
  } 
  break;
 }

 updateBackgroundImage();
}

void CoordinateSystem3D::setFineGrid(double newGridInterval, 
                                     bool   shouldBeVisible, int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS:
  {
   if( isLogScaledX )
   {
    jassert(newGridInterval > 1.00001);
     // for logarithmic scaling, we need the grid-intervals to be > 1
    if( newGridInterval <= 1.00001 )
    {
     fineGridIntervalX = 2.0;
     return;
    }
   }
   else
   {
    jassert(newGridInterval > 0.000001); 
     // grid-intervals must be > 0
    if( newGridInterval <= 0.000001 )
     return;
   }
   fineGridIsVisibleX = shouldBeVisible;
   fineGridIntervalX  = newGridInterval;
  } 
  break;
 case Y_AXIS:
  {
   if( isLogScaledY )
   {
    jassert(newGridInterval > 1.00001);
     // for logarithmic scaling, we need the grid-intervals to be > 1
    if( newGridInterval <= 1.00001 )
    {
     fineGridIntervalY = 2.0;
     return;
    }
   }
   else
   {
    jassert(newGridInterval > 0.000001); 
     // grid-intervals must be > 0
    if( newGridInterval <= 0.000001 )
     return;
   }
   fineGridIsVisibleY = shouldBeVisible;
   fineGridIntervalY  = newGridInterval;
  } 
  break;
 case Z_AXIS:
  {
   if( isLogScaledZ )
   {
    jassert(newGridInterval > 1.00001);
     // for logarithmic scaling, we need the grid-intervals to be > 1
    if( newGridInterval <= 1.00001 )
    {
     fineGridIntervalZ = 2.0;
     return;
    }
   }
   else
   {
    jassert(newGridInterval > 0.000001); 
     // grid-intervals must be > 0
    if( newGridInterval <= 0.000001 )
     return;
   }
   fineGridIsVisibleZ = shouldBeVisible;
   fineGridIntervalZ  = newGridInterval;
  } 
  break;
 }
 updateBackgroundImage();
}

bool CoordinateSystem3D::isCoarseGridVisible(int whichAxis) 
{ 
 switch( whichAxis )
 {
 case X_AXIS: return coarseGridIsVisibleX; 
 case Y_AXIS: return coarseGridIsVisibleY; 
 case Z_AXIS: return coarseGridIsVisibleZ; 
 default:     return false;
 }
}

bool CoordinateSystem3D::isFineGridVisible(int whichAxis) 
{ 
 switch( whichAxis )
 {
 case X_AXIS: return fineGridIsVisibleX; 
 case Y_AXIS: return fineGridIsVisibleY; 
 case Z_AXIS: return fineGridIsVisibleZ; 
 default:     return false;
 }
}

double CoordinateSystem3D::getCoarseGridInterval(int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS: return coarseGridIntervalX; 
 case Y_AXIS: return coarseGridIntervalY; 
 case Z_AXIS: return coarseGridIntervalZ; 
 default:     return 1.0;
 }
}

double CoordinateSystem3D::getFineGridInterval(int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS: return fineGridIntervalX; 
 case Y_AXIS: return fineGridIntervalY; 
 case Z_AXIS: return fineGridIntervalZ; 
 default:     return 1.0;
 }
}

void CoordinateSystem3D::setAxisPosition(int newAxisPosition, int whichAxis)
{
 switch( whichAxis )
 {
 case X_AXIS:
  {
   if( newAxisPosition == INVISIBLE ||
       newAxisPosition == ZERO      ||
       newAxisPosition == TOP       ||
       newAxisPosition == BOTTOM      )
   {
    axisPositionX = newAxisPosition;
   }
  } 
  break;
 case Y_AXIS:
  {
   if( newAxisPosition == INVISIBLE ||
       newAxisPosition == ZERO      ||
       newAxisPosition == LEFT      ||
       newAxisPosition == RIGHT        )
   {
    axisPositionY = newAxisPosition;
   }
  } 
  break;
 case Z_AXIS:
  {
   if( newAxisPosition == INVISIBLE ||
       newAxisPosition == ZERO      ||
       newAxisPosition == LEFT      ||
       newAxisPosition == RIGHT        )
   {
    axisPositionZ = newAxisPosition;
   }
  } 
  break;
 }
 updateBackgroundImage();
}

void CoordinateSystem3D::setRotationAboutAxis(double newRotation, 
 int whichAxis, bool unitIsDegrees)
{
 switch( whichAxis )
 {
 case X_AXIS:
  rotationAboutX = newRotation;
  break;
 case Y_AXIS:
  rotationAboutY = newRotation;
  break;
 case Z_AXIS:
  rotationAboutZ = newRotation;
  break;
 }
 updateProjectionParameters();
 updateBackgroundImage();
}

void CoordinateSystem3D::useStandardProjection(int newStandardProjection)
{
 projectionMethod = AXONOMETRIC;
 switch( newStandardProjection )
 {
 case ISOMETRIC:
  {
   rotationAboutY = -45.0;
   rotationAboutX =  asin(sqrt(1.0/3.0)) * 180.0/M_PI; 
  }
  break;
 case MILITARY:
  {
   rotationAboutY = -45.0;
   rotationAboutX =  45.0;
  }
  break;
 default: // isometric projection by default:
  {
   rotationAboutY = -45.0;
   rotationAboutX =  asin(sqrt(1.0/3.0)) * 180.0/M_PI; 
  }
 }
 updateProjectionParameters();
 updateBackgroundImage();
}

void CoordinateSystem3D::useLogarithmicScale(bool   shouldBeLogScaled, 
	int whichAxis, double newLogBase)
{
 switch( whichAxis )
 {
 case X_AXIS:
  {
   isLogScaledX = shouldBeLogScaled;
   logBaseX     = newLogBase;
  } 
  break;
 case Y_AXIS:
  {
   isLogScaledY = shouldBeLogScaled;
   logBaseY     = newLogBase;
  } 
  break;
 case Z_AXIS:
  {
   isLogScaledZ = shouldBeLogScaled;
   logBaseZ     = newLogBase;
  } 
  break;
 }
 updateProjectionParameters();
 updateBackgroundImage();
}

//----------------------------------------------------------------------------
void CoordinateSystem3D::paint(juce::Graphics &g)
{
 if( backgroundImage != NULL )
 {
  g.drawImage(backgroundImage, 0, 
                               0, 
                               getWidth(), 
                               getHeight(), 
                               0, 
                               0, 
                               backgroundImage->getWidth(), 
                               backgroundImage->getHeight(), 
                               false);
 }
 else
  g.fillAll(Colours::red);
}

void CoordinateSystem3D::updateBackgroundImage()
{
 if( getWidth() < 1 || getHeight() < 1 )
  return;

 // allocate memory for the first time:
 if( backgroundImage == NULL )
 {
  backgroundImage = new Image(Image::RGB, getWidth(), getHeight(), true);

  if( backgroundImage == NULL )
   return; // memory allocation failed
 }

 // reallocate memory, if necessary (i.e. the size of the component differs
 // from the size of the current image):
 if( backgroundImage->getWidth()  != getWidth()  ||
     backgroundImage->getHeight() != getHeight()    )
 {
  // delete the old and create a new Image-object:
  if( backgroundImage != NULL )
  {
   delete backgroundImage;
   backgroundImage = NULL;
  }
  backgroundImage = new Image(Image::RGB, getWidth(), getHeight(), true);

  if( backgroundImage == NULL )
   return; // memory allocation failed
 }

 // create a graphics object which is associated with the image to perform
 // the drawing-operations
 Graphics g(*backgroundImage);

 // fill the background:
 g.fillAll(backgroundColour);

 // draw the clipping planes, if desired:
 drawClippingPlane(g, XY_PLANE, clippingPlaneColourXY, 
  clippingPlanesOutlineColour);
 drawClippingPlane(g, XZ_PLANE, clippingPlaneColourXZ, 
  clippingPlanesOutlineColour);
 drawClippingPlane(g, YZ_PLANE, clippingPlaneColourYZ, 
  clippingPlanesOutlineColour);

 // draw the grids, if desired:
 if( fineGridIsVisibleX )
  drawGrid(g, fineGridIntervalX, X_AXIS, isLogScaledX, 
           fineGridColour, 1.0f);
 if( fineGridIsVisibleY)
  drawGrid(g, fineGridIntervalY, Y_AXIS, isLogScaledY, 
           fineGridColour, 1.0f);
 if( fineGridIsVisibleZ)
  drawGrid(g, fineGridIntervalZ, Z_AXIS, isLogScaledZ, 
           fineGridColour, 1.0f);
 if( coarseGridIsVisibleX )
  drawGrid(g, coarseGridIntervalX, X_AXIS, isLogScaledX, 
           coarseGridColour, 1.0f);
 if( coarseGridIsVisibleY )
  drawGrid(g, coarseGridIntervalY, Y_AXIS, isLogScaledY, 
           coarseGridColour, 1.0f);
 if( coarseGridIsVisibleZ )
  drawGrid(g, coarseGridIntervalZ, Z_AXIS, isLogScaledZ, 
           coarseGridColour, 1.0f);

 // draw the coordinate axes:
 if( axisPositionX != NO_ANNOTATION )
  drawAxis(g, X_AXIS);
 if( axisPositionY != NO_ANNOTATION )
  drawAxis(g, Y_AXIS);
 if( axisPositionZ != NO_ANNOTATION )
  drawAxis(g, Z_AXIS);

 // draw the labels on the axes:
 if( axisLabelPositionX != NO_ANNOTATION )
  drawAxisLabel(g, X_AXIS);
 if( axisLabelPositionY != NO_ANNOTATION )
  drawAxisLabel(g, Y_AXIS);
 if( axisLabelPositionZ != NO_ANNOTATION )
  drawAxisLabel(g, Z_AXIS);

 // draw the values on the axes:
 if( axisValuesPositionX != NO_ANNOTATION )
  drawAxisValues(g, X_AXIS);
 if( axisValuesPositionY != NO_ANNOTATION )
  drawAxisValues(g, Y_AXIS);
 if( axisValuesPositionZ != NO_ANNOTATION )
  drawAxisValues(g, Z_AXIS);

 // draw an outlining rectangle:
 g.setColour(outlineColour);
 g.drawRect(0, 0, getWidth(), getHeight(), 1);

 // trigger a repaint:
 repaint();
}

//----------------------------------------------------------------------------

void CoordinateSystem3D::transformToComponentsCoordinates(double &x, 
                                                          double &y,
                                                          double &z)
{
 /*
 double theta;
 double tmpX, tmpY, tmpZ;

 // scale all 3  coordinates to unit interval:
 x /= range3D.getMaxX() - range3D.getMinX();
 y /= range3D.getMaxY() - range3D.getMinY();
 z /= range3D.getMaxZ() - range3D.getMinZ();

 // rotate about y-axis (page 44, 1.30 b):
 theta = rotationAboutY * M_PI/180.0;
 tmpX  = x;
 tmpZ  = z;
 x     =  tmpX*cos(theta) + tmpZ*sin(theta);
 z     = -tmpX*sin(theta) + tmpZ*cos(theta);

 // rotate about x-axis (page 44, 1.30 c):
 theta = rotationAboutX * M_PI/180.0;
 tmpY  = y;
 tmpZ  = z;
 y     =  tmpY*cos(theta) - tmpZ*sin(theta);
 z     =  tmpY*sin(theta) + tmpZ*cos(theta);

 // rotate about z-axis (page 44, 1.30 a):
 theta = rotationAboutZ * M_PI/180.0;
 tmpX  = x;
 tmpY  = y;
 x     =  tmpX*cos(theta) - tmpY*sin(theta);
 y     =  tmpX*sin(theta) + tmpY*cos(theta);

 // scale to components witdh/height;
 x *= 400;
 y *= 400;
 //x *= 1.0*getWidth();
 //y *= 1.0*getHeight();

 // preliminary:
 x += 0.5*getWidth();
 y += 0.5*getHeight();

 // invert y:
 y = getHeight() - y;
 */


 // translate central-values to origin:
 x -= 0.5 * (range3D.getMinX() + range3D.getMaxX());
 y -= 0.5 * (range3D.getMinY() + range3D.getMaxY());
 z -= 0.5 * (range3D.getMinZ() + range3D.getMaxZ());


 // scale to unit interval:
 x /= (range3D.getMaxX()-range3D.getMinX());
 y /= (range3D.getMaxY()-range3D.getMinY());
 z /= (range3D.getMaxZ()-range3D.getMinZ());

 // calculate intermediate homogenous coordinates:
 double X = a*x + d*y + h*z + l;
 double Y = b*x + e*y + i*z + m;
 double Z = c*x + f*y + j*z + n;
 double H = p*x + q*y + r*z + s;

 // calculate final coordinates (by dividing by H) and assign them to the 
 // referenced variables:
 double oneOverH = 1.0 / H;
 x = oneOverH * X;
 y = oneOverH * Y;
 z = oneOverH * Z;

 // scale/shift by the user-adjustable parameters:
 x *= horizontalScale;
 y *= verticalScale;
 x -= horizontalShift;
 y -= verticalShift;


 // scale to components witdh/height;
 x *= 1.0*getWidth();
 y *= 1.0*getHeight();

 // preliminary:
 x += 0.5*getWidth();
 y += 0.5*getHeight();

 // invert y:
 y = getHeight() - y;

 int dummy = 1;
}

void CoordinateSystem3D::transformToComponentsCoordinates(float &x, 
                                                          float &y,
                                                          float &z)                                                         
{

}

//----------------------------------------------------------------------------

void CoordinateSystem3D::drawAxis(Graphics &g, int whichAxis)
{
 // check, if actually have to draw something, if not return immediately:
 switch( whichAxis )
 {
 case X_AXIS: if( axisPositionX == INVISIBLE ) return;  break;
 case Y_AXIS: if( axisPositionY == INVISIBLE ) return;  break;
 case Z_AXIS: if( axisPositionZ == INVISIBLE ) return;  break;
 default: return;
 }

 g.setColour(axesColour);
 double startX, endX, startY, endY, startZ, endZ;

 // read out start- and end-point of the axis to be drawn (in the 3 dimensional
 // coordinate-system):
 switch( whichAxis )
 {
 case X_AXIS:
  {
   startX = range3D.getMinX();
   endX	  = range3D.getMaxX();
   startY = 0.0;
   endY   = 0.0;
   startZ = 0.0;
   endZ   = 0.0;
  }
  break;
 case Y_AXIS:
  {
   startX = 0.0;
   endX	  = 0.0;
   startY = range3D.getMinY();
   endY   = range3D.getMaxY();
   startZ = 0.0;
   endZ   = 0.0;
  }
  break;
 case Z_AXIS:
  {
   startX = 0.0;
   endX	  = 0.0;
   startY = 0.0;
   endY   = 0.0;
   startZ = range3D.getMinZ();
   endZ   = range3D.getMaxZ();
  }
  break;
 }

 // transform to the two-dimensional coordinate-system of the component:
 transformToComponentsCoordinates(startX, startY, startZ);
 transformToComponentsCoordinates(endX, endY, endZ);

 // draw the axis as an arrow:
 g.drawArrow((float)startX, (float)startY, (float)endX, (float)endY, 
             2.0, 8.0, 8.0);
}

void CoordinateSystem3D::drawAxisLabel(Graphics &g, int whichAxis)
{
 // check, if actually have to draw something, if not return immediately:
 switch( whichAxis )
 {
 case X_AXIS: if( axisLabelPositionX == NO_ANNOTATION ) return;  break;
 case Y_AXIS: if( axisLabelPositionY == NO_ANNOTATION ) return;  break;
 case Z_AXIS: if( axisLabelPositionZ == NO_ANNOTATION ) return;  break;
 default: return;
 }

 g.setColour(axesColour);
 double posX, posY, posZ;
 String labelToDraw;

 // get the postion for the label in 3 dimensions:
 switch( whichAxis )
 {
 case X_AXIS:
  {
   posX	       = range3D.getMaxX();
   posY        = 0.0;
   posZ        = 0.0;
   labelToDraw = axisLabelX;
  }
  break;
 case Y_AXIS:
  {
   posX	       = 0.0;
   posY        = range3D.getMaxY();
   posZ        = 0.0;
   labelToDraw = axisLabelY;
  }
  break;
 case Z_AXIS:
  {
   posX	       = 0.0;
   posY        = 0.0;
   posZ        = range3D.getMaxZ();
   labelToDraw = axisLabelZ;
  }
  break;
 }

 // transform to the two-dimensional coordinate-system of the component:
 transformToComponentsCoordinates(posX, posY, posZ);

 // draw the label:
 g.drawText(labelToDraw, (int) posX+8, (int) posY, 64, 16, 
            Justification::centredLeft, false);
}

void CoordinateSystem3D::drawAxisValues(juce::Graphics &g, int whichAxis)
{
 // check, if actually have to draw something, if not return immediately:
 switch( whichAxis )
 {
 case X_AXIS: if( axisValuesPositionX == NO_ANNOTATION ) return;  break;
 case Y_AXIS: if( axisValuesPositionY == NO_ANNOTATION ) return;  break;
 case Z_AXIS: if( axisValuesPositionZ == NO_ANNOTATION ) return;  break;
 default: return;
 }

 // assign some local variables according to the axis to be drawn:
 bool   isLogScaled;
 double interval, minValue, maxValue;
 switch( whichAxis )
 {
 case X_AXIS:
  {
   isLogScaled  = isLogScaledX;
   interval     = coarseGridIntervalX;
   minValue     = range3D.getMinX();
   maxValue     = range3D.getMaxX();
  }
  break;
 case Y_AXIS:
  {
   isLogScaled  = isLogScaledY;
   interval     = coarseGridIntervalY;
   minValue     = range3D.getMinY();
   maxValue     = range3D.getMaxY();
  }
  break;
 case Z_AXIS:
  {
   isLogScaled  = isLogScaledZ;
   interval     = coarseGridIntervalZ;
   minValue     = range3D.getMinZ();
   maxValue     = range3D.getMaxZ();
  }
  break;
 } // end of switch( whichAxis )

 // check if the coarse-grid interval (which is the same as the interval 
 // between two values on the axis) is set to a reasonable value, otherwise 
 // return:
 if( isLogScaled == true )
 {
  jassert( interval >= 1.00001 );
  if( interval < 1.00001 )
   return;
 }
 else
 {
  jassert( interval >= 0.000001 );
   // grid spacing must be > 0
  if( interval < 0.000001 )
   return;
 }

 g.setColour(axesColour);
 int 	  i;
 double	x1, x2, y1, y2, z1, z2;
 double value = 0.0;
 //double accumulator;
 String numberString;

 // a function pointer for the appropriate function to convert the value into
 // a string:
 String (*stringConversionFunction) (double valueToConvert);
 stringConversionFunction = &valueToString0; // initializes the pointer

 if(isLogScaled)
 {


  // still something to do here....


 }
 else	 // axis is linerarly scaled
 {
  i = 0;
  while( i*interval < maxValue )
  {

   // determine coordinates at which the value has to drawn:
   switch( whichAxis )
   {
   case X_AXIS:
    {
     x1 = value = i*interval;
     y1 = range3D.getMinY();
     z1 = range3D.getMaxZ();
     x2 = x1;
     y2 = range3D.getMaxY();
     z2 = range3D.getMinZ();
     stringConversionFunction = stringConversionFunctionX;
    }
    break;
   case Y_AXIS:
    {
     x1 = range3D.getMaxX();
     y1 = value = i*interval;
     z1 = range3D.getMinZ();
     x2 = range3D.getMinX();
     y2 = value = y1;
     z2 = range3D.getMaxZ();
     stringConversionFunction = stringConversionFunctionY;
    }
    break;
   case Z_AXIS:
    {
     x1 = range3D.getMaxX();
     y1 = range3D.getMinY();
     z1 = value = i*interval;
     x2 = range3D.getMinX();
     y2 = range3D.getMaxY();
     z2 = value = z1;
     stringConversionFunction = stringConversionFunctionZ;
    }
    break;
   } // end of switch( whichAxis )

   // translate the number to be drawn into a string using the appropriate
   // stringConversionFunction:
   numberString = stringConversionFunction(value);

   // transform to the two-dimensional coordinate-system of the component:
   transformToComponentsCoordinates(x1, y1, z1); 
   transformToComponentsCoordinates(x2, y2, z2); 

   // draw the number:
   g.drawText(numberString, (int)x1-32, (int)y1-8, 64, 16, 
	     juce::Justification::centred, false);
   g.drawText(numberString, (int)x2-32, (int)y2-8, 64, 16, 
	     juce::Justification::centred, false);

   i++;

  } // end of while( i*interval < maxValue )
  i = 1;
  while( -i*interval > minValue )
  {

   // determine coordinates at which the value has to drawn:
   switch( whichAxis )
   {
   case X_AXIS:
    {
     x1 = value = -i*interval;
     y1 = range3D.getMinY();
     z1 = range3D.getMaxZ();
     x2 = x1;
     y2 = range3D.getMaxY();
     z2 = range3D.getMinZ();
     stringConversionFunction = stringConversionFunctionX;
    }
    break;
   case Y_AXIS:
    {
     x1 = range3D.getMaxX();
     y1 = value = -i*interval;
     z1 = range3D.getMinZ();
     x2 = range3D.getMinX();
     y2 = value = y1;
     z2 = range3D.getMaxZ();
     stringConversionFunction = stringConversionFunctionY;
    }
    break;
   case Z_AXIS:
    {
     x1 = range3D.getMaxX();
     y1 = range3D.getMinY();
     z1 = value = -i*interval;
     x2 = range3D.getMinX();
     y2 = range3D.getMaxY();
     z2 = value = z1;
     stringConversionFunction = stringConversionFunctionZ;
    }
    break;
   } // end of switch( whichAxis )

   // translate the number to be drawn into a string using the appropriate
   // stringConversionFunction:
   numberString = stringConversionFunction(value);

   // transform to the two-dimensional coordinate-system of the component:
   transformToComponentsCoordinates(x1, y1, z1); 
   transformToComponentsCoordinates(x2, y2, z2); 

   // draw the number:
   g.drawText(numberString, (int)x1-32, (int)y1-8, 64, 16, 
	     juce::Justification::centred, false);
   g.drawText(numberString, (int)x2-32, (int)y2-8, 64, 16, 
	     juce::Justification::centred, false);

   i++;

  } // end of while( -i*gridInterval > minValue )

 } // end of else	// axis is linerarly scaled

}

void CoordinateSystem3D::drawClippingPlane(Graphics &g, int whichPlane, 
 Colour planeColour, Colour outlineColour)
{
 Path   planeQuadrilateral;

 double x1, x2, x3, x4, y1, y2, y3, y4, z1, z2, z3, z4;

 switch( whichPlane )
 {
 case XY_PLANE:
  {
   x1 = range3D.getMinX();
   y1 = range3D.getMinY();
   z1 = range3D.getMinZ();
   transformToComponentsCoordinates(x1, y1, z1);

   x2 = range3D.getMaxX();
   y2 = range3D.getMinY();
   z2 = range3D.getMinZ();
   transformToComponentsCoordinates(x2, y2, z2);

   x3 = range3D.getMaxX();
   y3 = range3D.getMaxY();
   z3 = range3D.getMinZ();
   transformToComponentsCoordinates(x3, y3, z3);

   x4 = range3D.getMinX();
   y4 = range3D.getMaxY();
   z4 = range3D.getMinZ();
   transformToComponentsCoordinates(x4, y4, z4);

   planeQuadrilateral.addQuadrilateral((float) x1, (float) y1, 
    (float) x2, (float) y2, (float) x3, (float) y3, (float) x4, (float) y4);
  }
  break;
 case XZ_PLANE:
  {
   x1 = range3D.getMinX();
   y1 = range3D.getMinY();
   z1 = range3D.getMinZ();
   transformToComponentsCoordinates(x1, y1, z1);

   x2 = range3D.getMaxX();
   y2 = range3D.getMinY();
   z2 = range3D.getMinZ();
   transformToComponentsCoordinates(x2, y2, z2);

   x3 = range3D.getMaxX();
   y3 = range3D.getMinY();
   z3 = range3D.getMaxZ();
   transformToComponentsCoordinates(x3, y3, z3);

   x4 = range3D.getMinX();
   y4 = range3D.getMinY();
   z4 = range3D.getMaxZ();
   transformToComponentsCoordinates(x4, y4, z4);

   planeQuadrilateral.addQuadrilateral((float) x1, (float) y1, 
    (float) x2, (float) y2, (float) x3, (float) y3, (float) x4, (float) y4);
  }
  break;
 case YZ_PLANE:
  {
   x1 = range3D.getMinX();
   y1 = range3D.getMinY();
   z1 = range3D.getMinZ();
   transformToComponentsCoordinates(x1, y1, z1);

   x2 = range3D.getMinX();
   y2 = range3D.getMinY();
   z2 = range3D.getMaxZ();
   transformToComponentsCoordinates(x2, y2, z2);

   x3 = range3D.getMinX();
   y3 = range3D.getMaxY();
   z3 = range3D.getMaxZ();
   transformToComponentsCoordinates(x3, y3, z3);

   x4 = range3D.getMinX();
   y4 = range3D.getMaxY();
   z4 = range3D.getMinZ();
   transformToComponentsCoordinates(x4, y4, z4);

   planeQuadrilateral.addQuadrilateral((float) x1, (float) y1, 
    (float) x2, (float) y2, (float) x3, (float) y3, (float) x4, (float) y4);
  }
  break;
 }

 g.setColour(planeColour);
 g.fillPath(planeQuadrilateral);
 g.setColour(outlineColour);
 g.strokePath(planeQuadrilateral, PathStrokeType::curved);
 //g.strokePath(planeQuadrilateral, PathStrokeType::mitered);
  // strange: with mitered, it doesn't draw anything
}

void CoordinateSystem3D::drawGrid(Graphics &g, double interval, int whichAxis,
 bool exponentialSpacing, Colour gridColour, float lineThickness)
{
 //return;

 if( exponentialSpacing == true )
 {
  jassert( interval >= 1.00001 );
   // grid spacing must be > 1 for exponentially spaced grid-lines
  if( interval < 1.00001 )
   return;
 }
 else
 {
  jassert( interval >= 0.000001 );
   // grid spacing must be > 0
  if( interval < 0.000001 )
   return;
 }

 double minValue, maxValue;
 switch( whichAxis )
 {
 case X_AXIS:
  {
   minValue = range3D.getMinX();
   maxValue = range3D.getMaxX();
  }
  break;
 case Y_AXIS:
  {
   minValue = range3D.getMinY();
   maxValue = range3D.getMaxY();
  }
  break;
 case Z_AXIS:
  {
   minValue = range3D.getMinZ();
   maxValue = range3D.getMaxZ();
  }
  break;
 } // end of switch( whichAxis )


 g.setColour(gridColour);
 int 	  i;
 double x1, x2, x3, x4, y1, y2, y3, y4, z1, z2, z3, z4;
 //double accumulator;

 if(exponentialSpacing)
 {


  // still something to do here....


 }
 else	 // grid is linerarly scaled
 {
  i = 0;
  while( i*interval < maxValue )
  {
   // determine coordinates at which the value has to drawn:
   switch( whichAxis )
   {
   case X_AXIS:
    {
     x1 = i*interval;
     x2 = x1;
     y1 = range3D.getMinY();
     y2 = range3D.getMinY();
     z1 = range3D.getMinZ();
     z2 = range3D.getMaxZ();

     x3 = i*interval;
     x4 = x3;
     y3 = range3D.getMinY();
     y4 = range3D.getMaxY();
     z3 = range3D.getMinZ();
     z4 = range3D.getMinZ();
    }
    break;
   case Y_AXIS:
    {
     x1 = range3D.getMinX();
     x2 = range3D.getMaxX();
     y1 = i*interval;
     y2 = y1;
     z1 = range3D.getMinZ();
     z2 = range3D.getMinZ();

     x3 = range3D.getMinX();
     x4 = range3D.getMinX();
     y3 = i*interval;
     y4 = y3;
     z3 = range3D.getMinZ();
     z4 = range3D.getMaxZ();
    }
    break;
   case Z_AXIS:
    {
     x1 = range3D.getMinX();
     x2 = range3D.getMaxX();
     y1 = range3D.getMinY();
     y2 = range3D.getMinY();
     z1 = i*interval;
     z2 = z1;

     x3 = range3D.getMinX();
     x4 = range3D.getMinX();
     y3 = range3D.getMinY();
     y4 = range3D.getMaxY();
     z3 = i*interval;
     z4 = z3;
    }
    break;
   } // end of switch( whichAxis )

   // transform to the two-dimensional coordinate-system of the component:
   transformToComponentsCoordinates(x1, y1, z1); 
   transformToComponentsCoordinates(x2, y2, z2); 
   transformToComponentsCoordinates(x3, y3, z3); 
   transformToComponentsCoordinates(x4, y4, z4); 

   // draw two grid-lines:
   g.drawLine((float)x1, (float)y1, (float)x2, (float)y2, 1.0);
   g.drawLine((float)x3, (float)y3, (float)x4, (float)y4, 1.0);

   i++;
  }
  i = 1;
  while( -i*interval > minValue )
  {
   switch( whichAxis )
   {
   case X_AXIS:
    {
     x1 = -i*interval;
     x2 = x1;
     y1 = range3D.getMinY();
     y2 = range3D.getMinY();
     z1 = range3D.getMinZ();
     z2 = range3D.getMaxZ();

     x3 = -i*interval;
     x4 = x3;
     y3 = range3D.getMinY();
     y4 = range3D.getMaxY();
     z3 = range3D.getMinZ();
     z4 = range3D.getMinZ();
    }
    break;
   case Y_AXIS:
    {
     x1 = range3D.getMinX();
     x2 = range3D.getMaxX();
     y1 = -i*interval;
     y2 = y1;
     z1 = range3D.getMinZ();
     z2 = range3D.getMinZ();

     x3 = range3D.getMinX();
     x4 = range3D.getMinX();
     y3 = -i*interval;
     y4 = y3;
     z3 = range3D.getMinZ();
     z4 = range3D.getMaxZ();
    }
    break;
   case Z_AXIS:
    {
     x1 = range3D.getMinX();
     x2 = range3D.getMaxX();
     y1 = range3D.getMinY();
     y2 = range3D.getMinY();
     z1 = -i*interval;
     z2 = z1;

     x3 = range3D.getMinX();
     x4 = range3D.getMinX();
     y3 = range3D.getMinY();
     y4 = range3D.getMaxY();
     z3 = -i*interval;
     z4 = z3;
    }
    break;
   } // end of switch( whichAxis )

   // transform to the two-dimensional coordinate-system of the component:
   transformToComponentsCoordinates(x1, y1, z1); 
   transformToComponentsCoordinates(x2, y2, z2); 
   transformToComponentsCoordinates(x3, y3, z3); 
   transformToComponentsCoordinates(x4, y4, z4); 

   // draw two grid-lines:
   g.drawLine((float)x1, (float)y1, (float)x2, (float)y2, 1.0);
   g.drawLine((float)x3, (float)y3, (float)x4, (float)y4, 1.0);

   i++;
  }
 } // end of else	// axis is linerarly scaled
}

//-----------------------------------------------------------------------------

void CoordinateSystem3D::updateProjectionParameters()
{
 // calculate the rotation-matrix:
 double theta, c_x, s_x, c_y, s_y, c_z, s_z;

 theta = rotationAboutX * M_PI/180.0;
 c_x   = cos(theta);
 s_x   = sin(theta);

 theta = rotationAboutY * M_PI/180.0;
 c_y   = cos(theta);
 s_y   = sin(theta);

 theta = rotationAboutZ * M_PI/180.0;
 c_z   = cos(theta);
 s_z   = sin(theta);

 a =  c_y*c_z - s_y*s_x*s_z;
 b =  c_y*s_z + s_y*s_x*c_z;
 c = -s_y*c_x;

 d = -c_x*s_z;
 e =  c_x*c_z;
 f =  s_x;

 h = s_y*c_z + c_y*s_x*s_z;
 i = s_y*s_z - c_y*s_x*c_z;
 j = c_y*c_x;


 // include scaling-factors for the individual coordinate-axes into the 
 // rotation matrix:
 /*
 double sigma_x = 1.0 / (range3D.getMaxX() - range3D.getMinX());
 a *= sigma_x;
 b *= sigma_x;
 c *= sigma_x;

 double sigma_y = 1.0 / (range3D.getMaxY() - range3D.getMinY());
 d *= sigma_y;
 e *= sigma_y;
 f *= sigma_y;

 double sigma_z = 10. / (range3D.getMaxZ() - range3D.getMinZ());
 h *= sigma_z;
 i *= sigma_z;
 j *= sigma_z;
 */


 // calculate the translations:


 // assign the remaining parametrs:
 p = 0.0; 
 q = 0.0;
 r = 0.0;
 s = 1.0;

 /*
 scaleX = getWidth()  / (range3D.getMaxX()-range3D.getMinX()); 
 scaleY = getHeight() / (range3D.getMaxY()-range3D.getMinY());
 scaleZ = getWidth()  / (range3D.getMaxZ()-range3D.getMinZ());
 */
}

//-----------------------------------------------------------------------------
// state-management:

XmlElement* CoordinateSystem3D::getStateAsXml(const String& stateName) const
{
 XmlElement* xmlState = new XmlElement(stateName); 
  // the XmlElement which stores all the releveant state-information

 xmlState->setAttribute(String(T("MinX")), range3D.getMinX());
 xmlState->setAttribute(String(T("MaxX")), range3D.getMaxX());
 xmlState->setAttribute(String(T("MinY")), range3D.getMinY());
 xmlState->setAttribute(String(T("MaxY")), range3D.getMaxY());



 return xmlState;
}

bool CoordinateSystem3D::setStateFromXml(const XmlElement &xmlState)
{
 bool success = true; // should report about success, not used yet

 range3D.setMinX( xmlState.getDoubleAttribute(String(T("MinX")),
                  getMinimum(X_AXIS)) );
 range3D.setMaxX( xmlState.getDoubleAttribute(String(T("MaxX")),
                  getMaximum(X_AXIS)) );

 range3D.setMinX( xmlState.getDoubleAttribute(String(T("MinY")),
                  getMinimum(Y_AXIS)) );
 range3D.setMaxX( xmlState.getDoubleAttribute(String(T("MaxY")),
                  getMaximum(Y_AXIS)) );

 range3D.setMinX( xmlState.getDoubleAttribute(String(T("MinZ")),
                  getMinimum(Z_AXIS)) );
 range3D.setMaxX( xmlState.getDoubleAttribute(String(T("MaxZ")),
                  getMaximum(Z_AXIS)) );


 updateBackgroundImage();
 return success;
}




