/**

This class is similar to the CoordinateSystem class, except that it represents
a 3-dimensional coordinate-system. The axis naming convention is as follows:
the x-axis goes to the right, the y-axis goes upwards and the z-axis comes out 
of the screen - it is thus a so called right-handed coordinate-system. Right 
handed coordinate-systems are obtained by associating the right hand thumb 
with the x-axis, the index-finger with the y-axis and the middle finger with 
the z-axis. If you now rotate your hand such that the thumb points to the 
right, the index-finger points upwards and the middle-finger points at you, 
you have the coordinate-system which is represented by this class.

The internal implementation is largely based on the book "Transformations and 
Projections in Computer Graphics" by David Salomon and the mathematical 
notation used in this book is used here for variable-names - so if you want to
dig deeper into the implementation, refer to this excellent book.

*/

#ifndef __JUCE_COORDINATESYSTEM3D_JUCEHEADER__
#define __JUCE_COORDINATESYSTEM3D_JUCEHEADER__

#include "StringTools.h"
#include "Point3D.h"
#include "CoordinateSystem3DRange.h"

class CoordinateSystem3D : public Component
{

public:

 enum axes
 {
  X_AXIS,
  Y_AXIS,
  Z_AXIS
 };

 enum planes
 {
  XY_PLANE,
  XZ_PLANE,
  YZ_PLANE
 };

 enum axisPositions
 {
  INVISIBLE = 0,
  ZERO,
  LEFT,
  RIGHT,
  TOP,
  BOTTOM
 };

 enum axisAnnotationPositions
 {
  NO_ANNOTATION = 0,
  LEFT_TO_AXIS, 
  RIGHT_TO_AXIS,
  ABOVE_AXIS,
  BELOW_AXIS,
  AUTOMATIC
 };

 enum projectionMethods
 {
  AXONOMETRIC = 0,
  PERSPECTIVE
 };

 enum standardProjections
 {
  ISOMETRIC = 0,
  MILITARY
 };

 CoordinateSystem3D(const String &name = String(T("CoordinateSytem3D")));   
 ///< Constructor.

 virtual ~CoordinateSystem3D();            
 ///< Destructor.

 virtual void setBounds(int x, int y, int width, int height);
 ///< Overrides the setBounds-function of the component base-class.

 virtual void paint(Graphics &g);
 ///< Overrides the paint-function of the component base-class.

 virtual void setRange(double newMinX, double newMaxX, 
                       double newMinY, double newMaxY,
                       double newMinZ, double newMaxZ);
 /**< Sets the minimum and maximum values for the x-,y- and z-axis. For 
      logarithmic axis-scaling, make sure that the respective minimum value
      is greater than zero! */

 virtual void setRange(CoordinateSystem3DRange newRange);
 /**< Sets the currently visible range3D. */

 virtual CoordinateSystem3DRange getRange3D() { return range3D; }
 /**< Returns the currently visible range3D. */

 virtual void setMinimum(double newMinimum, int whichAxis);
 /**< Sets the minimum value for one of the 3 axes. */

 virtual double getMinimum(int whichAxis);
 /**< Returns the minimum value of one of the 3 axes. */

 virtual void setMaximum(double newMaximum, int whichAxis);
 /**< Sets the maximum value for one of the 3 axes. */

 virtual double getMaximum(int whichAxis);
 /**< Returns the maximum value of one of the 3 axes. */

 virtual CoordinateSystem3DRange 
  getMaximumMeaningfulRange3D(double relativeMargin = 10.0);
 /** Returns the maximum range3D which makes sense for a given plot as 
     CoordinateSystem3DRange-object. This is supposed to be overriden by 
     subclasses. */

 virtual void setAxisPosition(int newAxisPosition, int whichAxis);
 /**< Sets the position of the 3 axes. For possible values see 
      enum axisPositions. */

 virtual void setAxisLabel(const String &newLabel, int whichAxis,
                           int newLabelPosition = AUTOMATIC);
 /**< Sets the label for one of the 3 axes and its position. */

 virtual void setValuesPosition(int newValuesPosition, int whichAxis);
 /**< Switches value annotation postion for one of the 3 axes. */

 void setCoarseGrid(double newGridInterval, bool   shouldBeVisible, 
                    int whichAxis);
 /**< Sets the interval and visibility of the coarse grid for one of the 3 
      axes. */

 void setFineGrid(double newGridInterval, bool   shouldBeVisible, 
                  int whichAxis);
 /**< Sets the interval and visibility of the fine grid for one of the 3 
      axes. */

 virtual bool isCoarseGridVisible(int whichAxis);
 /**< Informs, if the coarse grid is visible for one of the 3 axes. */

 virtual bool isFineGridVisible(int whichAxis);
 /**< Informs, if the fine grid is visible for one of the 3 axes. */

 virtual double getCoarseGridInterval(int whichAxis);
 /**< Returns the interval of the coarse grid of one of the 3 axes. */

 virtual double getFineGridInterval(int whichAxis);
 /**< Returns the interval of the fine grid of one of the 3 axes. */

 void setRotationAboutAxis(double newRotation, int whichAxis, 
                           bool unitIsDegrees = true);
 /**< Sets the rotation about one of the 3 axes. For objects that move into 
      positive z-direction (out of the screen, towards you), these rotations 
      are also known as pitch, yaw and roll where pitch is rotation about the
      x-axis, yaw is rotation about the y-axis and roll is rotation about the 
      z-axis. */

 void useStandardProjection(int newStandardProjection);
 /**< Sets up the rotation angles about y- and x-axis according to certain
      common standards for axonometric projections. */

 //virtual void setClippingPlanePosition(double newClippingPlanePosition,
 //                                      int whichPlane);
 /**< Sets the position of the clipping-plane which is parallel to a given 
      plane. For each given plane, this is always the remaining 
      coordinate, so for the clipping plane parallel to the xy-plane, 
      it is the z-value for example.  */

 //void setClippingPlaneVisible(bool shouldBeVisible, int whichPlane);
 /**< Makes one of the 3 clipping-planes visible or not. */

 virtual void useLogarithmicScale(bool   shouldBeLogScaled, int whichAxis,
                                  double newLogBase = 2.0);
 /**< Decides, if an axis should be logarithmically scaled and sets up the
      base for the logarithm. */

 virtual void setStringConversionFunction(
  String (*newConversionFunction) (double valueToBeConverted), int whichAxis);
 /**< This function is used to pass a function-pointer. This pointer has to be
      the address of a function which has a double-parameter and a juce::String
      as return-value. The function will be used to convert the values on one 
      of the 3 axes into corresponding strings for display. */

 virtual void setColourScheme(Colour newBackgroundColour, 
                              Colour newAxesColour,
                              Colour newCoarseGridColour, 
                              Colour newFineGridColour,
                              Colour newOutLineColour);
 /**< Sets up the colour-scheme for the CoordinateSystem. */


 virtual void setHorizontalScaleFactor(double newScaleFactor);
 /**< Sets up a scale (zoom-) factor for the horizontal component-coordinate. */

 virtual double getHorizontalScaleFactor();
 /**< Returns the scale (zoom-) factor for the horizontal component-coordinate. */

 virtual void setVerticalScaleFactor(double newScaleFactor);
 /**< Sets up a scale (zoom-) factor for the vertical component-coordinate. */

 virtual double getVerticalScaleFactor();
 /**< Returns the scale (zoom-) factor for the vertical component-coordinate. */

 virtual void setHorizontalShift(double newShift);
 /**< Sets up a left-/right for the horizontal screen-coordinates (in normalized 
      coordinates, where 1 corresponds to the component-width). */

 virtual double getHorizontalShift();
 /**< Returns the left-/right for the horizontal screen-coordinates (in normalized 
      coordinates, where 1 corresponds to the component-width). */

 virtual void setVerticalShift(double newShift);
 /**< Sets up a up-/down for the vertical screen-coordinates (in normalized 
      coordinates, where 1 corresponds to the component-width). */

 virtual double getVerticalShift();
 /**< Returns the up-/down for the horizontal screen-coordinates (in normalized 
      coordinates, where 1 corresponds to the component-width). */

 virtual void scaleHorizontalScaleFactor(double scaler);
 /**< Scales the horizontal scale-factor by multiplying it with scaler. */


 virtual void scaleVerticalScaleFactor(double scaler);
 /**< Scales the vertical scale-factor by multiplying it with scaler. */

 virtual void incrementHorizontalShift(double increment);
 /**< Increments the horizontal shift by adding the increment (which can be also 
      negative in which case it results in a decrement). */

 virtual void incrementVerticalShift(double increment);
 /**< Increments the vertical shift by adding the increment (which can be also 
      negative in which case it results in a decrement). */

 virtual void incrementRotationAngle(double increment, int whichAxis);
 /**< Increments the rotation angle about one of the three axes. */

 //void setProjectionMethod(int newProjectionMethod);
 /**< Chooses the method by which points in 3D-space will be projected onto 
      points in 2D-space. See enum projectionMethods for possible values. */

 //void setCameraPosition(Point3D newCameraPosition);
 /**< Sets up the position of the virtual camera which looks at the scene. */

 //void setCameraDirection(Point3D newCameraDirection);
 /**< Sets up the direction into which the virtual camera looks. */

 virtual XmlElement* getStateAsXml(
  const String& stateName = String(T("CoordinateSystem3DState"))) const;
 /**< Creates an XmlElement from the current state and returns it. */

 virtual bool setStateFromXml(const XmlElement &xmlState);
 /**< Restores a state based on an XmlElement which should have been created
      with the getStateAsXml()-function. */

protected:

 CoordinateSystem3DRange range3D;
 /**< The range3D-object for the coordinate system. */

 // the setup-variables:
 double scaleX, scaleY, scaleZ;
 double pixelsPerIntervalX, pixelsPerIntervalY, pixelsPerIntervalZ;
 int    axisPositionX, axisPositionY, axisPositionZ;
 String axisLabelX, axisLabelY, axisLabelZ;
 int    axisLabelPositionX, axisLabelPositionY, axisLabelPositionZ;
 int    axisValuesPositionX, axisValuesPositionY, axisValuesPositionZ;
 bool   coarseGridIsVisibleX, fineGridIsVisibleX, 
        coarseGridIsVisibleY, fineGridIsVisibleY,
        coarseGridIsVisibleZ, fineGridIsVisibleZ;
 double coarseGridIntervalX, fineGridIntervalX, 
        coarseGridIntervalY, fineGridIntervalY, 
        coarseGridIntervalZ, fineGridIntervalZ;
 bool   isLogScaledX, isLogScaledY, isLogScaledZ;
 double logBaseX, logBaseY, logBaseZ;

 int    projectionMethod;
 double rotationAboutX, rotationAboutY, rotationAboutZ;

 double verticalScale, horizontalScale;
 double verticalShift, horizontalShift;
  // scaling and shifting of the complete picture (after being projected, 
  // but before being scaled to components width/height)

 double maxScale, minScale, maxShift, minShift;
  // maximum and minimum values for the parameters above


 // the transformation-parameters (see page 37 in the book metioned in the 
 // class-description for their meaning):
 double a, b, c, p,
        d, e, f, q,
        h, i, j, r,
        l, m, n, s;

 // the colour-scheme:
 Colour backgroundColour;
 Colour axesColour;
 //Colour clippingPlanesColour;
 Colour clippingPlaneColourXY;
 Colour clippingPlaneColourXZ;
 Colour clippingPlaneColourYZ;
 Colour clippingPlanesOutlineColour;
 Colour coarseGridColour;
 Colour fineGridColour;
 Colour outlineColour;

 String  headline;
 /**< Meant for a plot-headline - not used yet. */

 Image*  backgroundImage; 
  // this image will be used for the appearance of the coodinate system, it 
  // will be updated via the updateBackgroundImage()-function when something
  // about the coordinate-system (axes, grid, etc.) changes

 // functions for drawing the background (for internal use):
 virtual void drawAxis(Graphics &g, int whichAxis);
 virtual void drawAxisLabel(Graphics &g, int whichAxis);
 virtual void drawAxisValues(Graphics &g, int whichAxis);

 virtual void drawClippingPlane(Graphics &g, int whichPlane, 
               Colour planeColour, Colour outlineColour);

 //virtual void drawGridOnClippingPlane(Graphics &g, int whichPlane, 
 //              double interval, bool exponentialSpacing, Colour gridColour,
 //              float lineThickness = 1.f);

 virtual void drawGrid(Graphics &g, double interval, int whichAxis,
  bool exponentialSpacing, Colour gridColour, float lineThickness);






 virtual void transformToComponentsCoordinates(double &x, double &y, 
                                               double &z);
 /**< Function for converting the x- and y-coordinate values into the 
      corresponding coordinates in the component (double precision version).*/

 virtual void transformToComponentsCoordinates(float &x, float &y, float &z);
 /**< Function for converting the x- and y-coordinate values into the 
      corresponding coordinates in the component (single precision version).*/

 virtual void updateBackgroundImage();
 /**< Updates the image object (re-draws it). Will be called, when something
      about the CoordinateSystem3D's appearance-settings was changed. */

 virtual void updateProjectionParameters();
 /**< Updates the scale-factors which are needed when transforming from the 
      CoordinateSystem3D's coordinates to Component's coordinates and vice 
      versa. Will be called by setBounds(), setRange() and 
      useLogarithmicScale(). */

 String (*stringConversionFunctionX) (double valueToConvert);
 /**< A pointer to the function which converts a x-value into a juce-string. */

 String (*stringConversionFunctionY) (double valueToConvert);
 /**< A pointer to the function which converts a y-value into a juce-string. */

 String (*stringConversionFunctionZ) (double valueToConvert);
 /**< A pointer to the function which converts a z-value into a juce-string. */

};

#endif  // __JUCE_COORDINATESYSTEM3D_JUCEHEADER__
