#include "FileManager.h"

//----------------------------------------------------------------------------
// construction/destruction:

FileManager::FileManager() : fileFilter(String(T("*")), String(T("All Files")))
{
  fileIsUnsaved           = false;
  currentFileName         = String::empty;
  currentFileNameWithStar = String::empty;
  currentFileFullPath     = String::empty;
  currentFileRelativePath = String::empty;
  wildcardPatterns        = String(T("*"));
  defaultExtension        = String::empty;

  updateFileList();
}

FileManager::~FileManager()
{

}

//----------------------------------------------------------------------------
// 

void FileManager::openLoadingDialog(const String &dialogTitle)
{
  FileChooser chooser(dialogTitle, File(currentDirectory), wildcardPatterns, true);

  if(chooser.browseForFileToOpen())
  {
    File fileToLoad = chooser.getResult();

    // remember where we are:
    currentFileFullPath     = fileToLoad.getFullPathName();
    currentDirectory        = fileToLoad.getParentDirectory().getFullPathName(); 
    currentFileRelativePath = fileToLoad.getRelativePathFrom(File::getSpecialLocation(
      File::currentExecutableFile));

    // update the preset-field on the GUI
    currentFileName         = fileToLoad.getFileName();
    currentFileNameWithStar = currentFileName + String(T("*"));

    // update the array with the presets which are in the current 
    // preset-directory (which is needed for preset increment/decrement):
    updateFileList();
  }
}

void FileManager::openSavingDialog(const String &dialogTitle)
{
  FileChooser chooser(dialogTitle, File(currentDirectory), wildcardPatterns, true);

  if(chooser.browseForFileToSave(true))
  {
    File fileToSaveTo = chooser.getResult();

    // auto append the the extension .xml if the user did not type it in
    // manually:
    if ( !fileToSaveTo.hasFileExtension(defaultExtension) && defaultExtension != String::empty )
      fileToSaveTo = fileToSaveTo.withFileExtension( defaultExtension ) ;

    // remember where we are:
    currentFileFullPath     = fileToSaveTo.getFullPathName();
    currentDirectory        = fileToSaveTo.getParentDirectory().getFullPathName(); 
    currentFileRelativePath = fileToSaveTo.getRelativePathFrom(File::getSpecialLocation(
      File::currentExecutableFile));

    // update the preset-field on the GUI
    currentFileName         = fileToSaveTo.getFileName();
    currentFileNameWithStar = currentFileName + String(T("*"));

    // update the array with the presets which are in the current 
    // preset-directory (which is needed for preset increment/decrement):
    updateFileList();
  }
}

void FileManager::updateFileList()
{
  if(File(currentDirectory).exists() && File(currentDirectory).isDirectory())
  {
    // update the array with the presets which are in the current 
    // preset-directory (which is needed for preset increment/decrement):
    filesInCurrentDirectory.clear();

    /*
    File(currentDirectory).findChildFiles(filesInCurrentDirectory, 
      File::findFiles, 
      false, 
      wildcardPatterns);
    */
    // if we want to match one from a set of different wildcard-pattern, we have to do it manually:

    // find all files in this directory:
    File(currentDirectory).findChildFiles(filesInCurrentDirectory, File::findFiles, false);

    // remove all files from the array, which do not match the require wildcard-pattern (we can't
    // simply use the wildcardPattern parameter of the findChildFiles-function when we look for
    // several different wildcard-patterns - this works only for one single pattern):
    int fileIndexToCheck = 0;
    while( fileIndexToCheck < filesInCurrentDirectory.size() )
    {
      if( !fileFilter.isFileSuitable(*filesInCurrentDirectory[fileIndexToCheck]) )
      {
        filesInCurrentDirectory.remove(fileIndexToCheck, true);
        // this will also take care of decrementing filesInCurrentDirectory.size() - that's why we
        // use a while-loop instead of just iterating through all indices of whole array in its
        // initial size
      }
      fileIndexToCheck++;
    }

    numFilesInCurrentDirectory = filesInCurrentDirectory.size();
    filesInCurrentDirectory.sort(fileComparator);

    // find the index of the currently loaded file inside the current directory:
    currentFileIndex = 0;
    for(int i=0; i<numFilesInCurrentDirectory; i++)
    {
      if(filesInCurrentDirectory[i]->getFileName() == currentFileName)
      {
        currentFileIndex = i;
        break;
      }
    }
  }
}

void FileManager::incrementCurrentFile()
{
  // make sure that we are not in an empty directory:
  if(numFilesInCurrentDirectory <= 0)
    return;

  // increment the index with wraparound:
  currentFileIndex += 1;
  while( currentFileIndex > numFilesInCurrentDirectory-1 )
    currentFileIndex -= numFilesInCurrentDirectory;

  File* fileToLoadFrom = filesInCurrentDirectory[currentFileIndex];
  if(fileToLoadFrom != NULL &&
    fileToLoadFrom->existsAsFile() )
  {
    currentFileFullPath     = fileToLoadFrom->getFullPathName();
    currentFileName         = fileToLoadFrom->getFileName();
    currentFileNameWithStar = currentFileName + String(T("*"));
    currentFileRelativePath = fileToLoadFrom->getRelativePathFrom(File::getSpecialLocation(
      File::currentExecutableFile));
  }
}

void FileManager::decrementCurrentFile()
{
  // make sure that we are not in an empty directory:
  if(numFilesInCurrentDirectory <= 0)
    return;

  // increment the index with wraparound:
  currentFileIndex -= 1;
  while( currentFileIndex < 0 )
    currentFileIndex += numFilesInCurrentDirectory;

  File* fileToLoadFrom = filesInCurrentDirectory[currentFileIndex];
  if(fileToLoadFrom != NULL &&
    fileToLoadFrom->existsAsFile() )
  {
    currentFileFullPath     = fileToLoadFrom->getFullPathName();
    currentFileName         = fileToLoadFrom->getFileName();
    currentFileNameWithStar = currentFileName + String(T("*"));
    currentFileRelativePath = fileToLoadFrom->getRelativePathFrom(File::getSpecialLocation(
      File::currentExecutableFile));
  }
}

File FileManager::getCurrentFile()
{
  return File(currentFileFullPath);
}

void FileManager::retrieveCurrentDirectoryFromPreferences(String xmlAttributeToLookFor)
{
  // retrieve current application-directory:
  File   thisExeAsFile         = File::getSpecialLocation(File::currentExecutableFile);
  File   thisDirectoryAsFile   = thisExeAsFile.getParentDirectory();
  String thisDirectoryAsString = thisDirectoryAsFile.getFullPathName();

  // look for a 'Preferences.xml' file:
  File preferencesFile = File(thisDirectoryAsString + String(T("/PreferencesGeneral.xml")) );

  if( preferencesFile.existsAsFile() )
  {
    XmlDocument myDocument(preferencesFile);
    XmlElement* xmlPreferences = myDocument.getDocumentElement();

    if( xmlPreferences != NULL )
    {
      String parentDirectory = xmlPreferences->getStringAttribute(
        String(T("ContentDirectoryAbsolute")), String(T("ApplicationDirectory")) );

      String subDirectory = xmlPreferences->getStringAttribute(
        xmlAttributeToLookFor, String::empty );

      if( parentDirectory == String(T("ApplicationDirectory")) )
        currentDirectory = thisDirectoryAsString + subDirectory;
      else
        currentDirectory = parentDirectory + subDirectory;

      delete xmlPreferences;
    }
  }
}

