/**

This class is a component, which can be used to plot a surface in some
underlying 3 dimensional coordinate-system. 

*/

#include "CoordinateSystem3D.h"

#ifndef __JUCE_SURFACEPLOT_JUCEHEADER__
#define __JUCE_SURFACEPLOT_JUCEHEADER__

class SurfacePlot : public CoordinateSystem3D
{

public:

  enum gridModes
  {
    TRIANGULAR = 0,
    QUADRANGULAR
  };

  SurfacePlot(const String& name = String(T("SurfacePlot")));   
  ///< Constructor.

  virtual ~SurfacePlot();   
  ///< Destructor.

  virtual void paint(Graphics &g);
  ///< Overrides the paint-function of the component base-class.

  virtual void setSurfaceValues(int      newNumValueX1, 
    double*  newValuesX, 
    int      newNumValuesX2,
    double*  newValuesX2,
    double** newValuesY);
  /**<  Accepts the new function or curve values to be drawn.
  The newValuesX1 and newValuesX1 are assumed to arrays of length 
  newNumValuesX1 and newNumValuesX2 respectively - they represent the 
  grid at which the two independent variables are sampled. The 
  2-dimensional array nwValuesY is assumed to be an array of size 
  newNumValuesX1 times newNumValuesX2 and represents the dependent 
  variable (the height above the x1, x2 - plane).

  NOTE:
  The data will NOT be copied into internal buffers, an
  object of class "SurfacePlot" will only hold pointers to the data. If
  the outlying class for some reason deletes the arrays in which the data
  is stored, it should call invalidatePointers() - such that the 
  "SurfacePlot"-object does not try to aceess the array anymore. */

  virtual void invalidatePointers();
  /**< Sets the pointers to the arrays which hold the function values to NULL.
  Should be called, if some outlying class deletes the arrays in which the
  x- and y-values are stored - remember that the SurfacePlot-class 
  doesn't store the arrays by itself but accesses them via pointers which
  must be set correctly by the outlying class which uses the 
  SurfacePlot-class. */

  virtual void setDrawAsCloud(bool shouldBeDrawnAsCloud)
  { drawAsCloud = shouldBeDrawnAsCloud; }
  /**< Values should be drawn as dots like in a 3D - scatter plot. */

  virtual void setSurfaceColor(Colour newSurfaceColour)
  { surfaceColour = newSurfaceColour; }
  /**< Selects the colour for the actual surface. */

  virtual void setMeshGridColor(Colour newMeshGridColour)
  { meshGridColour = newMeshGridColour; }
  /**< Selects the colour for the mesh-grid. */

protected:

  // functions for drawing the data itself:
  virtual void plotSurface(Graphics &g, bool plotWithMeshGrid = true, 
    bool plotWithDots = false);
  virtual void plotCloud(Graphics &g);

  // pointers to the data (and integers for the array-sizes):
  int      numValuesX1;
  int      numValuesX2;
  double*  valuesX1;
  double*  valuesX2;
  double** valuesY;

  Colour   surfaceColour, meshGridColour, dotColour;

  bool     drawAsCloud;  // draw the function values as dots without 
  // connecting them with a line

  bool     plotSurfaceWithMesh;
  bool     plotSurfaceWithDots;

  int      gridMode;
};

#endif  // __JUCE_SURFACEPLOT_JUCEHEADER__
