#ifndef rosic_ExpressionEvaluatorFunctions_h
#define rosic_ExpressionEvaluatorFunctions_h

/**

This file defines custom functions (and mathematical constants) for the ExpressionEvaluator.  

*/

// standard library includes:
#include <errno.h>
extern int errno;  // is set in functions from math.h and can be checked here

// third party includes:
#include "ExprEval_v3_4/expreval.h"
using namespace ExprEval;
using namespace std;

// rosic-indcludes:
#include "rosic_RealFunctions.h"
using namespace rosic;



//-------------------------------------------------------------------------------------------------
// Gaussian (aka normal) distribution:

class FunctionNodeGauss : public FunctionNode
{
public:
  FunctionNodeGauss(Expression *expression) : FunctionNode(expression)
  { SetArgumentCount(1, 3, 0, 0, 0, 0);  }

  double DoEvaluate()
  {
    if( m_nodes.size() < 2 )
      return gauss(m_nodes[0]->Evaluate(), 0.0, 1.0); 
    else if( m_nodes.size() < 3 )
      return gauss(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate(), 1.0); 
    else
      return gauss(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate(), m_nodes[2]->Evaluate()); 
  }
};

class FunctionFactoryGauss : public FunctionFactory
{
public:
  std::string GetName() const                    { return "gauss"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodeGauss(expression);  }
};

//-------------------------------------------------------------------------------------------------
// Chebychev polynomial of arbitrary order:

class FunctionNodeCheby : public FunctionNode
{
public:
  FunctionNodeCheby(Expression *expression) : FunctionNode(expression)
  { SetArgumentCount(2, 2, 0, 0, 0, 0);  }

  double DoEvaluate()
  { return cheby(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate()); }
};

class FunctionFactoryCheby : public FunctionFactory
{
public:
  std::string GetName() const                    { return "cheby"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodeCheby(expression);  }
};

//-------------------------------------------------------------------------------------------------
// power function:

class FunctionNodePow : public FunctionNode
{
public:
  FunctionNodePow(Expression *expression) : FunctionNode(expression)
  { SetArgumentCount(2, 2, 0, 0, 0, 0);  }

  double DoEvaluate()
  { return pow(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate() ); }
};

class FunctionFactoryPow : public FunctionFactory
{
public:
  std::string GetName() const                    { return "pow"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodePow(expression);  }
};

//-------------------------------------------------------------------------------------------------
// bipolar power function:

class FunctionNodePowBipolar : public FunctionNode
{
public:
  FunctionNodePowBipolar(Expression *expression) : FunctionNode(expression)
  { SetArgumentCount(2, 2, 0, 0, 0, 0);  }

  double DoEvaluate()
  { return powBipolar(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate() ); }
};

class FunctionFactoryPowBipolar : public FunctionFactory
{
public:
  std::string GetName() const                    { return "powBipolar"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodePowBipolar(expression);  }
};

//-------------------------------------------------------------------------------------------------
// quantization to an arbitrary interval:

class FunctionNodeQuant : public FunctionNode
{
public:
  FunctionNodeQuant(Expression *expression) : FunctionNode(expression)
  { SetArgumentCount(2, 2, 0, 0, 0, 0);  }

  double DoEvaluate()
  { return quant(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate()); }
};

class FunctionFactoryQuant : public FunctionFactory
{
public:
  std::string GetName() const                    { return "quant"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodeQuant(expression);  }
};

//-------------------------------------------------------------------------------------------------
// quantization to some number of bits:

class FunctionNodeQuantToBits : public FunctionNode
{
public:
  FunctionNodeQuantToBits(Expression *expression) : FunctionNode(expression)
  { SetArgumentCount(2, 2, 0, 0, 0, 0);  }

  double DoEvaluate()
  { return quantToBits(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate()); }
};

class FunctionFactoryQuantToBits : public FunctionFactory
{
public:
  std::string GetName() const                    { return "quantToBits"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodeQuantToBits(expression);  }
};

//-------------------------------------------------------------------------------------------------
// sawtooth function:

class FunctionNodeSaw : public FunctionNode
{

public:

  FunctionNodeSaw(Expression *expression) : FunctionNode(expression)
  {
    SetArgumentCount(
      1,               // minimum number of normal arguments
      1,               // maximum number of normal arguments
      0,               // minimum number of reference arguments
      0,               // maximum number of reference arguments
      0,               // minimum number of data arguments
      0                // maximum number of data arguments
      );
  }

  double DoEvaluate()
  {
    return sawWave( m_nodes[0]->Evaluate() );
  }

};

class FunctionFactorySaw : public FunctionFactory
{
public:
  std::string GetName() const                    { return "saw"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodeSaw(expression);  }
};


//-------------------------------------------------------------------------------------------------
// signum function:

class FunctionNodeSign : public FunctionNode
{
public:
  FunctionNodeSign(Expression *expression) : FunctionNode(expression)
  { SetArgumentCount(1, 1, 0, 0, 0, 0);  }

  double DoEvaluate()
  { return sign(m_nodes[0]->Evaluate() ); }
};

class FunctionFactorySign : public FunctionFactory
{
public:
  std::string GetName() const                    { return "sign"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodeSign(expression);  }
};


//-------------------------------------------------------------------------------------------------
// soft-clipping:

class FunctionNodeSoftClip : public FunctionNode
{
public:
  FunctionNodeSoftClip(Expression *expression) : FunctionNode(expression)
  { SetArgumentCount(1, 5, 0, 0, 0, 0);  }

  double DoEvaluate()
  {
    if( m_nodes.size() < 2 )
      return softClip(m_nodes[0]->Evaluate()); 
    else if( m_nodes.size() < 3 )
      return softClip(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate()); 
    else if( m_nodes.size() < 4 )
      return softClip(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate(), m_nodes[2]->Evaluate()); 
    else if( m_nodes.size() < 5 )
      return softClip(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate(), m_nodes[2]->Evaluate(), 
                      m_nodes[3]->Evaluate()); 
    else
      return softClip(m_nodes[0]->Evaluate(), m_nodes[1]->Evaluate(), m_nodes[2]->Evaluate(), 
                      m_nodes[3]->Evaluate(), m_nodes[4]->Evaluate()); 

  }
};

class FunctionFactorySoftClip : public FunctionFactory
{
public:
  std::string GetName() const                    { return "softClip"; }
  FunctionNode* DoCreate(Expression *expression) { return new FunctionNodeSoftClip(expression);  }
};





#endif  // rosic_ExpressionEvaluatorFunctions_h