#ifndef rosic_HelperFunctions_h
#define rosic_HelperFunctions_h

// standard includes:
#include <math.h>

// rosic includes:
#include "rosic_Definitions.h"

namespace rosic
{

  INLINE int bitReverse(int number, int numBits);
  /**< Returns an integer that is represented by the bit-reversed input number, taking into account
  only 'numBits' bits in the reversal process. */

  template <class T>
  INLINE void orderBitReversedOutOfPlace(T *inBuffer, T *outBuffer, int length, int numBits);
  /**< Rearranges/permutes and array of class T into bit-reversed order. The 'length' MUST be the
  'numBits' th power of two (this is not checked for). */

  //template <class T>
  //INLINE void orderBitReversedInPlace(T* buffer, int length, int numStages);
  /**< Rearranges/permutes and array of class T into bit-reversed order (in place). The 'length' 
  MUST be the 'numStages' th power of two (this is not checked for). */

  template <class T>
  INLINE T product(T *buffer, int length);
  /**< Returns the product of the elements in the buffer for classes which define the 
  multiplication operator (the *= version thereof) and a constructor which can take an int 
  paramater as argument and initializes to the multiplicative neutral element of that class when 1 
  is passed . */

  template <class T>
  INLINE void swap(T &in1, T &in2);
  /**< Swaps two objects of class T. */

  template <class T>
  INLINE T sum(T *buffer, int length);
  /**< Returns the sum of the elements in the buffer for classes which define the 
  addition operator (the += version thereof) and a constructor which can take an int 
  paramater as argument and initializes to the additive neutral element of that class when 0 
  is passed . */


  //===============================================================================================
  // implementation fo the inlie functions:


  INLINE int bitReverse(int number, int numBits)
  {
    int result = 0;
    for(int n=0; n<numBits; n++)
    {
      // leftshift the previous result by one and accept the new LSB of the current number on the 
      // right:
      result   = (result << 1) + (number & 1);

      // rightshift the number to make the second bit from the right to the new LSB:
      number >>= 1;
    }
    return result;
  }

  template <class T>
  INLINE void orderBitReversedOutOfPlace(T *inBuffer, T *outBuffer, int length, int numBits)
  {
    // gather up the values from the bit-reversed positions in the input array:
    for(int n=0; n<length; n++)
      outBuffer[n] = inBuffer[bitReverse(n, numBits)];
  }

  template <class T>
  INLINE T product(T *buffer, int length)
  {
    T accu = T(1); // assume that a constructor call with 1 as argument initilizes the accu to
                   // the multiplicative neutral element of the class T.

    // multiplicatively accumulate:
    for(int n=0; n<length; n++)
      accu *= buffer[n];

    return accu;
  }

  template <class T>
  INLINE void rosic::swap(T &in1, T &in2)
  {
    T tmp = in1;
    in1   = in2;
    in2   = tmp;   
  }

  template <class T>
  INLINE T sum(T *buffer, int length)
  {
    T accu = T(0); // assume that a constructor call with 0 as argument initilizes the accu to
                   // the additive neutral element of the class T.

    // accumulate:
    for(int n=0; n<length; n++)
      accu += buffer[n];

    return accu;
  }

  /*
  template <class T>
  INLINE void rosic::orderBitReversedInPlace(T *buffer, int length, int numStages)
  {

  }
  */


} // end namespace rosic

#endif // #ifndef rosic_HelperFunctions_h