// maybe the command-line option /FORCE:MULTIPLE has to be added
// to the linker options

#ifndef rosic_RealFunctions_h
#define rosic_RealFunctions_h

// standard library includes:
//#include <emmintrin.h> // is this standard or microsoft-specific?
#include <math.h>
#include <stdlib.h>

// rosic includes:
#include "rosic_CephesDeclarations.h" 
#include "rosic_RealFunctionEvaluationAlgorithms.h"
#include "rosic_HelperFunctions.h"

namespace rosic
{

  INLINE double amp2dB(double amp);
  ///< Converts a raw amplitude value/factor to a value in decibels.

  INLINE double amp2dBWithCheck(double amp, double lowAmplitude = 0.000001);
  /**< Converts a raw amplitude value/factor to a value in decibels with a check, if the 
       amplitude is close to zero (to avoid log-of-zero and related errors). */

  INLINE double acosh(double x);
  /**< Inverse hyperbolic cosine. */

  INLINE double acoshs(double x);
  /**< Symmetrized inverse hyperbolic cosine. */

  INLINE double asinh(double x);
  /**< Inverse hyperbolic sine. */

  template <class T>
  INLINE int arrayMaxIndex(T* theArray, int numValues);
  //INLINE int arrayMaxIndex(double* doubleArray, int numValues);
  /**< Returns the index of the maximum value in an array of doubles where the
  array should be of length numValues. */

  template <class T>
  INLINE int arrayMinIndex(T* theArray, int numValues);
  //INLINE int arrayMinIndex(double* doubleArray, int numValues);
  /**< Returns the index of the minimum value in an array of doubles where the
  array should be of length numValues. */

  INLINE double belowOrAbove(double x, double low, double high);
  /**< Returns -1.0 if x is below low, 0.0 if x is between low and high and
  1.0 if x is above high. */   

  INLINE double beatsToSeconds(double beat, double bpm);
  /**< Converts a time-stamp given in beats into seconds acording to a tempo 
  measured in beats per minute (bpm). */

  // bessel function n */
  //INLINE double besselj(double order, double x);

  INLINE double cheby(double x, double n);
  ///< Returns the value of the n-th chebychev polynomial.

  INLINE double ceil(double x);
  ///< Returns the next greater integer number.

  INLINE double centerClip(double x, double t);
  ///< Center clips the x at threshold t.

  INLINE double clip(double x, double min, double max);
  ///< Clips x at the threshold t.

  INLINE double cd(double u, double m);
  /**< Returns the value of the Jacobian elliptic function cd(u,m) with argunment u and 
  parameter m ( m=k^2 with k being elliptic modulus (?) ) . */

  INLINE double cn(double u, double m);
  /**< Returns the value of the Jacobian elliptic function cn(u,m) with argunment u and 
  parameter m ( m=k^2 with k being elliptic modulus (?) ) . */

  INLINE double dB2amp(double x);
  ///< Converts a value in decibels to a raw amplitude value/factor (cost: 180 cycles).

  INLINE double dn(double u, double m);
  /**< Returns the value of the Jacobian elliptic function dn(u,m) with argunment u and 
  parameter m ( m=k^2 with k being elliptic modulus (?) ) . */

  INLINE int ellipticFunctions(double u, double m, 
    double *sn, double *cn, double *dn, double* phi);
  /**< Evaluates the Jacobian elliptic functions sn(u|m), cn(u|m), and dn(u|m) of parameter m 
  between 0 and 1, and real argument u. phi is the amplitdue of u. */

  void ellipticIntegral(double k, double *K, double *Kprime, int M=7);
  /**< Evaluates the complete elliptic integral of the first kind with elliptic modulus k. K will 
  be assigned to the quarter period K(k) and Kprime will be assigned to the quarter period 
  K'(k) = K(k'), k' = sqrt(1-k^2). M is the number of Landen iterations which determines the 
  precision of the calculation. */

  INLINE double exp10(double x);
  ///< Calculates the exponential function with base 10.

  INLINE double exp2(double x);
  ///< Calculates the exponential function with base 2.

  INLINE double foldOver(double x, double min, double max);
  ///< foldover at the specified value */

  INLINE double frac(double x);
  ///< Returns the fractional part of x.

  INLINE double freqToPitch(double freq);
  ///< Converts a frequency in Hz into a MIDI-note value.

  // gamma function
  //double gamma(double x);

  INLINE double gauss(double x, double mu, double sigma);
  ///< Gaussian function with mean mu and standard deviation sigma.

  INLINE bool isEven(int x);
  INLINE bool isOdd(int x);
  INLINE bool isPowerOfTwo(unsigned int x);

  template <class T>
  INLINE T limit(T x, T lowerBound, T upperBound);
  /**< limits the value of an object. */

  INLINE double log10(double x);
  ///< Calculates the logarithm to base 10.

  INLINE double log2(double x);
  ///< Calculates the logarithm to base 2.

  INLINE double logB(double x, double b);
  ///< Calculates logarithm to an arbitrary base b.

  INLINE double logistic(double x, double b);
  ///< Calculates the logistic function with slope parameter b.

  INLINE void makeBlackmanWindow(double *window, int length);
  /**< Fills the window-array with a Hamming-window. */

  INLINE void makeCosinePowerWindow(double *window, int length, double power = 2.0);
  /**< Fills the window-array with a cosine power window. */

  INLINE void makeHammingWindow(double *window, int length);
  /**< Fills the window-array with a Hamming-window. */

  INLINE void makeHanningWindow(double *window, int length);
  /**< Fills the window-array with a Hanning-window. */

  INLINE void makeRectangularWindow(double *window, int length);
  /**< Fills the window-array with a rectangular window. */


  INLINE double mapLinearToLinear(double in, double inMin, double inMax,
    double outMin, double outMax);
  /**< Converts a value between inMin and inMax into a value between outMin
  and outMax where the mapping is linear for the input and the output.
  an example: y = mapLinearToLinear(x, 0.0, 1.0, -96.0, 24.0) will map
  the input x assumed to lie inside 0.0...1.0 to the range between
  -96.0...24.0. This function is useful to convert between parameter
  representations between 0.0...1.0 and the clear-text parameters. */

  INLINE double mapLinearToExponential(double in, double inMin, double inMax,
    double outMin, double outMax);
  /**< Converts a value between inMin and inMax into a value between outMin
  and outMax where the mapping of the output is exponential.
  an example: y = mapLinearToLinear(x, 0.0, 1.0, 20.0, 20000.0) will map
  the input x assumed to lie inside 0.0...1.0 to the range between
  20.0...20000.0 where equal differences in the input lead to equal 
  factors in the output. Make sure that the outMin value is greater than 
  zero! */

  INLINE double mapExponentialToLinear(double in, double inMin, double inMax,
    double outMin, double outMax);
  /**< The Inverse of "mapLinearToExponential" */

  template <class T>
  INLINE T max(T in1, T in2);
  /**< The maximum of two objects on which the ">"-operator is defined. */

  template <class T>
  INLINE T max(T in1, T in2, T in3, T in4);
  /**< The maximum of four objects on which the ">"-operator is defined. */

  template <class T>
  INLINE T min(T in1, T in2);
  /**< The minimum of two objects on which the ">"-operator is defined. */

  template <class T>
  INLINE T min(T in1, T in2, T in3, T in4);
  /**< The minimum of four objects on which the ">"-operator is defined. */

  INLINE double minkowski(double x, double mu, double sigma, double k);
  /**< Minkowski distribution - a generalization of the gaussian distribution,
  where the exponent k inside the exponential is an addtional parameter.
  For k=2, it reduces to the well known gaussian distribution. */

  template <class T>
  INLINE T nextPowerOfTwo(T x);
  /**< Returns a power of two which is greater than or equal to the input 
  argument. */

  INLINE double pitchOffsetToFreqFactor(double pitchOffset);
  /**< Converts a picth-offset in semitones value into a frequency 
  multiplication factor. */

  INLINE double pitchToFreq(double pitch);
  ///< Converts a MIDI-note value into a frequency in Hz (cost: 75 cycles).

  INLINE double powBipolar(double base, double exponent);
  /**< Calculates the power of the absolute-value of some number and re-applies 
  its original sign afterwards, if base==0, the result will be 0 also. */

  INLINE int powInt(int base, int exponent);
  /**< Power function for intergers. Multiplies the base exponent-1 time with itself and 
  returns the result. */

  INLINE double quant(double x, double i);
  ///< Quantizes x to the interval i.

  INLINE double quantToBits(double x, double nBits);
  ///< Quantizes x (assumed to be in -1.0 - 1.0) to a grid which represents a resolution of nBits bits.

  INLINE double random(double min=0.0, double max=1.0);
  /**< Generates a pseudo-random number between min and max. */

  INLINE double randDbl();
  ///< Generates a random number between -1.0...+1.0 in double format.

  INLINE double round(double x);
  ///< Returns the nearest integer (as double).

  INLINE int roundToInt(double x);
  ///< Returns the nearest integer.

  //INLINE int roundToIntSSE2(double x);
  ///< Returns the nearest integer.

  INLINE double sawWave(double x);
  ///< Generates a 2*pi periodic saw wave.

  INLINE double secondsToBeats(double timeInSeconds, double bpm);
  ///< Converts a time value in seconds into a time value measured in beats.

  INLINE double sign(double x);
  ///< Returns the sign of x as double.

  INLINE void sinCos(double x, double* sinResult, double* cosResult);
  /**< Calculates sine and cosine of x more efficiently than calling sin(x) 
  and cos(x) seperately. */

  INLINE double sn(double u, double m);
  /**< Returns the value of the Jacobian elliptic function sn(u,m) with argunment u and 
  parameter m ( m=k^2 with k being elliptic modulus (?) ) . */

  INLINE void sinCosApprox(double x, double* sinResult, double* cosResult);
  /**< Calculates a parabolic approximation of the sine and cosine of x. */

  INLINE double softClip(double x, double lowClamp   = -1.0, double highClamp  =  1.0,
                                   double lowKnee    =  0.1, double highKnee   =  0.1 );
  /**< Applies a softclipping with a tanh shaped transtion region between the linear part and the
  clipping value. */

  INLINE double sqrWave(double x);
  ///< Generates a 2*pi periodic square wave.

  INLINE double srem(double x, double y);
  /**< Same syntax as fmod but brings the result into the symmetric interval -y/2...y/2 */

  INLINE double step(double x);
  ///< The unit step function - returns 1 for x>=0, 0 else.

  INLINE double triWave(double x);
  ///< Generates a 2*pi periodic triangle wave.

  INLINE double wholeNotesToSeconds(double noteValue, double bpm);
  /**< Converts a time-stamp given in whole notes into seconds according to a tempo measured in 
  beats per minute (bpm). */

  // wrapAround 


  //============================================================================
  //implementation:

  INLINE double acosh(double x)
  {
    if( x >= 1 )
      return log(x + sqrt(x*x-1) );
    else
    {
      DEBUG_BREAK; // argument must be >= 1, for the general case use the complex version acosh_c
                   // defined in rosic_ComplexFunctions.h
      return 0.0;
    }
  }

  INLINE double acoshs(double x)
  {
    if( fabs(x) >= 1 )
      return log(fabs(x + sqrt(x*x-1)) );
    else
    {
      DEBUG_BREAK; // fabs(x) must be >= 1, for the general case use the complex version acosh_c
                   // defined in rosic_ComplexFunctions.h
      return 0.0;
    }
  }


  INLINE double rosic::amp2dB(double amp)
  {
    return 8.6858896380650365530225783783321 * log(amp);
    //return 20*log10(amp); // naive version
  }

  INLINE double rosic::amp2dBWithCheck(double amp, double lowAmplitude)
  {
    if( amp >= lowAmplitude )
      return 8.6858896380650365530225783783321 * log(amp);
    else
      return 8.6858896380650365530225783783321 * log(lowAmplitude);
  }

  INLINE double asinh(double x)
  {
    return log(x + sqrt(x*x+1) );
  }

  template <class T>
  INLINE int rosic::arrayMaxIndex(T* theArray, int numValues)
  {
    int    maxIndex = 0;
    double maxValue = theArray[0];

    for(int i=0; i<numValues; i++)
    {
      if( theArray[i] > maxValue )
      {
        maxValue = theArray[i];
        maxIndex = i;
      }
    }

    return maxIndex;
  }

  template <class T>
  INLINE int rosic::arrayMinIndex(T* theArray, int numValues)
  {
    int    minIndex = 0;
    double minValue = theArray[0];

    for(int i=0; i<numValues; i++)
    {
      if( theArray[i] < minValue )
      {
        minValue = theArray[i];
        minIndex = i;
      }
    }

    return minIndex;
  }

  INLINE double rosic::beatsToSeconds(double beat, double bpm)
  {
    return (60.0/bpm)*beat;
  }

  INLINE double rosic::belowOrAbove(double x, double low, double high)
  {
    if( x < low )
      return -1.0;
    else if ( x > high )
      return 1.0;
    else
      return 0.0;
  }

  /*
  INLINE double rosic::besselj(double order, double x)
  {
  double vm;
  double *jv, *yv, *djv, *dyv;

  bessjyv(order, x, vm, jv, yv, djv, dyv);

  return *jv;
  }
  */

  INLINE double rosic::cheby(double x, double n)
  {
    if( fabs(x) <= 1.0 )
      return cos( n * acos(x) );
    else if( x > 0.0 ) 
      return cosh( n * acosh(x) );
    else
      return pow(-1.0, roundToInt(n) ) * cosh( n * acosh(-x) );
  }

  INLINE double rosic::ceil(double x)
  {
    //double frac = x - floor(x);
    return x + (1-frac(x));  
  }

  INLINE double rosic::clip(double x, double min, double max)
  {
    if( x > max )
      return max;
    else if ( x < min )
      return min;
    else return x;
  }

  INLINE int ellipticFunctions(double u, double m,             
    double *sn, double *cn, double *dn, double* phi)
  {
    return ellpj(u, m, sn, cn, dn, phi);
  }

  INLINE double rosic::cd(double u, double m)
  {
    double cdResult, snDummy, cnDummy, dnDummy, phiDummy;
    ellipticFunctions(u, m, &snDummy, &cnDummy, &dnDummy, &phiDummy);
    cdResult = cnDummy/dnDummy;
    return cdResult;
  }

  INLINE double rosic::cn(double u, double m)
  {
    double snDummy, cnResult, dnDummy, phiDummy;
    ellipticFunctions(u, m, &snDummy, &cnResult, &dnDummy, &phiDummy);
    return cnResult;
  }

  INLINE double rosic::dB2amp(double dB)
  {
    return exp(dB * 0.11512925464970228420089957273422);
    //return pow(10.0, (0.05*dB)); // naive, inefficient version
  }

  INLINE double rosic::dn(double u, double m)
  {
    double snDummy, cnDummy, dnResult, phiDummy;
    ellipticFunctions(u, m, &snDummy, &cnDummy, &dnResult, &phiDummy);
    return dnResult;
  }

  INLINE double rosic::exp10(double x)
  {
    return exp(LN10*x);
  }

  INLINE double rosic::exp2(double x)
  {
    return exp(LN2*x);
  }



  INLINE double rosic::foldOver(double x, double min, double max)
  {
    if( x > max )
      return max - (x-max);
    else if( x < min )
      return min - (x-min);
    else return x;
  }

  INLINE double rosic::frac(double x)
  {
    return x - floor(x);
  }

  INLINE double rosic::freqToPitch(double freq)
  {
    return 12.0 * log2(freq/440.0) + 69.0;
  }

  INLINE double rosic::gauss(double x, double mu, double sigma)
  {
    return (1.0 / (sqrt(2*PI)*sigma)) * exp( - ((x-mu)*(x-mu)) / (2*sigma*sigma) );
  }

  INLINE bool rosic::isEven(int x)
  {
    if( x%2 == 0 )
      return true;
    else
      return false;
  }

  INLINE bool rosic::isOdd(int x)
  {
    if( x%2 != 0 )
      return true;
    else
      return false;
  }

  INLINE bool rosic::isPowerOfTwo(unsigned int x)
  {
    unsigned int currentPower    = 1;

    while( currentPower <= x )
    {
      if( currentPower == x )
        return true;

      currentPower *= 2;
    }

    return false;
  }

  template <class T>
  INLINE T limit(T x, T lowerBound, T upperBound)
  {
    if( x < lowerBound )
      return lowerBound;
    if( x > upperBound )
      return upperBound;
    return x;
  }

  INLINE double rosic::log10(double x)
  {
    return ONE_OVER_LN10*log(x);
  }

  INLINE double rosic::log2(double x)
  {
    return ONE_OVER_LN2*log(x);
  }

  INLINE double rosic::logB(double x, double b)
  {
    return log(x)/log(b);
  }

  INLINE double rosic::logistic(double x, double b)
  {
    return 1.0 / ( 1 + exp(-b*x) );
  }

  INLINE void rosic::makeBlackmanWindow(double *window, int length)
  {
    for(int n=0; n<length; n++)
      window[n] = 0.42 - 0.5*cos( 2.0*PI*n / (double) (length-1)) 
                       + 0.08*cos(4.0*PI*n / (double) (length-1)) ;
  }

  INLINE void rosic::makeCosinePowerWindow(double *window, int length, double power)
  {
    for(int n=0; n<length; n++)
      window[n] = pow( sin(PI * (double) n / (double) length), power );
  }

  INLINE void rosic::makeHammingWindow(double *window, int length)
  {
    for(int n=0; n<length; n++)
      window[n] = 0.54 -  0.46 * cos(2.0*PI*n / (double) (length-1)) ;
  }

  INLINE void rosic::makeHanningWindow(double *window, int length)
  {
    for(int n=0; n<length; n++)
      window[n] = 0.5 * ( 1.0 - cos(2.0*PI*n / (double) (length-1)) );
  }

  INLINE void rosic::makeRectangularWindow(double *window, int length)
  {
    for(int n=0; n<length; n++)
      window[n] = 1.0;
  }

  INLINE double rosic::mapLinearToLinear(double in, double inMin, double inMax,
    double outMin, double outMax)
  {
    double tmp;

    // map input to the range 0.0...1.0:
    tmp = (in-inMin) / (inMax-inMin);

    // map the tmp-value to the range outMin...outMax:
    tmp *= (outMax-outMin);
    tmp += outMin;

    return tmp;
  }

  INLINE double rosic::mapLinearToExponential(double in, 
    double inMin, double inMax,
    double outMin, double outMax)
  {
    double tmp;

    // map input to the range 0.0...1.0:
    tmp = (in-inMin) / (inMax-inMin);

    // map the tmp-value exponentially to the range outMin...outMax:
    //tmp = outMin * exp( tmp*(log(outMax)-log(outMin)) );
    tmp = outMin * exp( tmp*(log(outMax/outMin)) );

    return tmp;
  }

  INLINE double rosic::mapExponentialToLinear(double in, 
    double inMin, double inMax,
    double outMin, double outMax)
  {
    double tmp;

    tmp = log(in/inMin) / log(inMax/inMin);

    tmp = outMin + tmp * (outMax-outMin);

    return tmp;
  }

  template <class T>
  INLINE T rosic::max(T in1, T in2)
  {
    if( in1 > in2 )
      return in1;
    else
      return in2;
  }

  template <class T>
  INLINE T rosic::max(T in1, T in2, T in3, T in4)
  {
    return max(max(in1, in2), max(in3, in4));
  }

  template <class T>
  INLINE T rosic::min(T in1, T in2)
  {
    if( in1 < in2 )
      return in1;
    else
      return in2;
  }

  template <class T>
  INLINE T rosic::min(T in1, T in2, T in3, T in4)
  {
    return min(min(in1, in2), min(in3, in4));
  }


  /*
  INLINE double rosic::max(double in1, double in2)
  {
  if( in1 >= in2 )
  return in1;
  else
  return in2;
  }

  INLINE int rosic::max(int in1, int in2)
  {
  if( in1 >= in2 )
  return in1;
  else
  return in2;
  }
  */

  INLINE double rosic::minkowski(double x, double mu, double sigma, double k)
  {
    return (1 /(sqrt(2*PI)*sigma)) * exp( - ( pow(abs(x-mu),k) ) / (2*pow(sigma,k)) );
    //return (1 /(sqrt(2*PI*pow(sigma,k)))) * exp( - ( pow(abs(x-mu),k) ) / (2*pow(sigma,k)) );
  }

  template <class T>
  INLINE T rosic::nextPowerOfTwo(T x)
  {
    T accu = 1;
    while(accu < x)
      accu *= 2;
    return accu;
  }

  INLINE double rosic::pitchOffsetToFreqFactor(double pitchOffset)
  {
    return exp(0.057762265046662109118102676788181 * pitchOffset);
    //return pow(2.0, pitchOffset/12.0); // naive, slower but numerically more precise
  }

  INLINE double rosic::pitchToFreq(double pitch)
  {
    return 8.1757989156437073336828122976033 * exp(0.057762265046662109118102676788181*pitch);
    //return 440.0*( pow(2.0, (pitch-69.0)/12.0) ); // naive, slower but numerically more precise
  }

  INLINE double rosic::powBipolar(double base, double exponent)
  {
    if( base > 0.0 )
      return pow(base, exponent);
    else if( base < 0.0 )
      return sign(base) * pow(fabs(base), exponent);
    else
      return 0.0;
  }

  INLINE int powInt(int base, int exponent)
  {
    int result = base;
    for(int p=1; p<exponent; p++)
      result *= base;
    return result;
  }

  INLINE double rosic::quant(double x, double i)
  {
    if(i <= DBL_MIN)
    {
      DEBUG_BREAK;
      return x;
    }

    double absX  = fabs(x);
    double signX = sign(x);
    double tmp   = fmod(x, i);
    if(x>=0)
    {
      if(tmp < 0.5*i)
        return x - tmp;
      else
        return x - tmp + i;
    }
    else
    {
      if(tmp < -0.5*i)
        return x - tmp - i;
      else
        return x - tmp;
    }
  }

  INLINE double rosic::quantToBits(double x, double nBits)
  {
    double interval = 2 / pow(2.0, nBits);
    return rosic::quant(x, interval);
  }

  INLINE double rosic::randDbl()
  {
    return (  ( (double)rand() - 16383.5 ) / 16383.5 );
  }

  INLINE double random(double min, double max)
  {
    double tmp = rand() / RAND_MAX;  // between 0...1
    return mapLinearToLinear(tmp, 0.0, 1.0, min, max);
  }

  INLINE double rosic::round(double x)
  {
    if( x-floor(x) >= 0.5 )
      return ceil(x);
    else
      return floor(x);
  }

  INLINE int rosic::roundToInt(double const x) 
  {
    int n;
#if defined(__unix__) || defined(__GNUC__)
    // 32-bit Linux, Gnu/AT&T syntax:
    __asm ("fldl %1 \n fistpl %0 " : "=m"(n) : "m"(x) : "memory" );
#else
    // 32-bit Windows, Intel/MASM syntax:
    __asm fld qword ptr x;
    __asm fistp dword ptr n;
#endif
    return n;
  }

  INLINE double rosic::secondsToBeats(double timeInSeconds, double bpm)
  {
    return timeInSeconds*(bpm/60.0);
  }

  /*
  INLINE int rosic::roundToIntSSE2(double const x) 
  {
    return _mm_cvtsd_si32(_mm_load_sd(&x));
  }
  */

  INLINE double rosic::sign(double x)
  {
    if(x<0)
      return -1.0;
    else if(x>0)
      return 1.0;
    else
      return 0;
  }

  INLINE void rosic::sinCos(double x, double* sinResult, double* cosResult)
  {

    // this is the assembly-version:
    double s, c;
    __asm fld x
      __asm fsincos
      __asm fstp c
      __asm fstp s
      *sinResult = s;
    *cosResult = c;

    /*
    // this is the version, which fuzzes around with the square-root:
    *sinResult = sin(x);
    if( x < 0.5*PI )
    *cosResult = sqrt(1 - (*sinResult) * (*sinResult));
    else if ( x < (3.0/2.0)*PI )
    *cosResult = -sqrt(1 - (*sinResult) * (*sinResult));
    else
    *cosResult = sqrt(1 - (*sinResult) * (*sinResult));
    */

    /*
    // this is the unoptimized standard version:
    *sinResult = sin(x);
    *cosResult = cos(x);
    */
  }

  INLINE void rosic::sinCosApprox(double x, double* sinResult, 
    double* cosResult)
  {
    static const double c = 0.70710678118654752440;

    // restrict input x to the range 0.0...2*PI:
    while( x > 2.0*PI )
      x -= 2*PI;
    while( x < 0.0 )
      x += 2*PI;

    if( x < PI/2 )
    {
      double tmp1 = x;
      double tmp2 = (2/PI) * tmp1 - 0.5;
      double tmp3 = (2-4*c)*tmp2*tmp2 + c;
      *sinResult  = tmp3 + tmp2;
      *cosResult  = tmp3 - tmp2;
    }
    else if( x < PI )
    {
      double tmp1 = (x-PI/2);
      double tmp2 = 0.5 - (2/PI) * tmp1;
      double tmp3 = (2-4*c)*tmp2*tmp2 + c;
      *sinResult  = tmp2 + tmp3;
      *cosResult  = tmp2 - tmp3;
    }
    else if( x < 1.5*PI )
    {
      double tmp1 = (x-PI);
      double tmp2 = (2/PI) * tmp1 - 0.5;
      double tmp3 = (4*c-2)*tmp2*tmp2 - c;
      *sinResult  = tmp3 - tmp2;
      *cosResult  = tmp3 + tmp2; 
    }
    else
    {
      double tmp1 = (x-1.5*PI);
      double tmp2 = (2/PI) * tmp1 - 0.5;
      double tmp3 = (2-4*c)*tmp2*tmp2 + c;
      *sinResult  = tmp2 - tmp3;
      *cosResult  = tmp2 + tmp3;  
    }

  }

  INLINE double rosic::sn(double u, double m)
  {
    double snResult, cnDummy, dnDummy, phiDummy;
    ellipticFunctions(u, m, &snResult, &cnDummy, &dnDummy, &phiDummy);
    return snResult;
  }

  INLINE double rosic::sawWave(double x)
  {
    double tmp = fmod(x, 2*PI);
    if( tmp < PI )
      return tmp/PI;
    else
      return (tmp/PI)-2.0;
  }

  INLINE double rosic::sqrWave(double x)
  {
    double tmp = fmod(x, 2*PI);
    if( tmp < PI )
      return 1.0;
    else
      return -1.0;
  }

  INLINE double softClip(double x,  double lowClamp, double highClamp,
                                    double lowKnee,  double highKnee)
  {
    double mid        = 0.5*(highClamp+lowClamp);
    double high       = highClamp-mid;
    double low        = lowClamp-mid;
    double highThresh = mid + (1.0-highKnee)*high;
    double lowThresh  = mid + (1.0-lowKnee)*low;

    // we want to deal with the absolute values of the thresholds and clamp values:
    lowThresh = mid-lowThresh;
    lowClamp  = mid-lowClamp;

    double highAlpha = highClamp - highThresh;
    if(highAlpha == 0.0)
      highAlpha += DBL_MIN;
    double highBeta = 1.0/highAlpha;

    double lowAlpha = lowClamp - lowThresh;
    if(lowAlpha == 0.0)
      lowAlpha += DBL_MIN;
    double lowBeta = 1.0/lowAlpha;

    double out;
    if(x >= mid)     // positive half-wave
    {
      if(x <= highThresh)
        out = x;
      else
        out = highThresh + highAlpha*tanh(highBeta*(x-highThresh));
    }
    else              // negative half-wave
    {
      x = mid-x;
      if(x <= lowThresh)
        out = x;
      else
        out = lowThresh + lowAlpha*tanh(lowBeta*(x-lowThresh));
      out = mid-out;
    }

    return out;
  }

  INLINE double rosic::srem(double x, double y)
  {
    double z = fmod(x, y);

    if( abs(z) > 0.5*y )
      z = z - y*sign(z);

    return z;
  }

  INLINE double rosic::triWave(double x)
  {
    double tmp = fmod(x, 2*PI);
    if( tmp < 0.5*PI )
      return tmp/(0.5*PI);
    else if( tmp < 1.5*PI )
      return 1.0 - ((tmp-0.5*PI)/(0.5*PI));
    else
      return -1.0 + ((tmp-1.5*PI)/(0.5*PI));
  }

  INLINE double rosic::wholeNotesToSeconds(double noteValue, double bpm)
  {
    return (240.0/bpm)*noteValue;
  }


} // end namespace rosic

#endif // #ifndef rosic_RealFunctions_h